"""Functions used for creating a Boolean mask from one or a sequence of images.

The resulting masks can then be used
for manipulating a :class:`~vcams.voxelpart.VoxelPart` instance
using its :meth:`~vcams.voxelpart.VoxelPart.apply_mask` method.
See the :ref:`predefined-image` section for a complete explanation
of the basic concepts.
"""

from logging import getLogger
from warnings import warn

import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap
from numpy import moveaxis, unique, ndarray, where, rot90
from numpy.core.numeric import array_equal
from skimage.filters.thresholding import threshold_otsu
from skimage.io import imread, ImageCollection
from skimage.restoration import denoise_bilateral
from skimage.transform import rescale

logger = getLogger(__name__)

# A list of available thresholding modes used in the GUI.
image_thresh_mode_list = ('none', 'manual', 'otsu')


def mask_from_image(image_path: str, scale: float = 1.0,
                    denoise: bool = True, show_image: bool = False,
                    thresh_mode: str = 'otsu', thresh_value: float = None,
                    verbose_log: bool = True, final_log: bool = True) -> ndarray:
    """Return a boolean mask by thresholding an image.

    This function does the following in the given order:

      1. Scale the image. ``skimage.transform.rescale()`` is used with ``anti_aliasing=True``.
      2. If specified, Denoise the image using ``skimage.restoration.denoise_bilateral()``.
      3. Apply a threshold using ``skimage.filters.threshold_otsu()``.
      4. If specified, show the opened image as grayscale and the final binary image.

    Args:
        image_path: Full path to the image file. The image will be opened as grayscale.
        scale: Scale to be applied to the image. Note that a scale greater than 1.0
               will introduce fake precision by interpolating the data and issues a warning.
        denoise: If set to True, the image will be denoised using a Bilateral filter.
        thresh_mode: The kind of threshold to be applied to the image.
                     Valid values are 'none', 'manual', and 'otsu'. Defaults to 'otsu'.
        thresh_value: If *thresh_mode* is set to 'manual', should be a float in the range (0, 1)
                      Which is used as a threshold for binarizing the image.
                      Defaults to *None* which raises an error when used.
        show_image: If set to True, the opened image and the final binary image
                    will be shown side by side in a figure.
                    The program may be paused while the window is open.
        verbose_log: If set to True, logs will be verbose, otherwise only one entry will be made.
        final_log: If set to True, the final log entry will be made.

    Returns:
        The binary mask derived from the image.
    """
    if scale <= 0.0:
        raise ValueError('scale must be positive.')
    if scale > 1.0:
        warn('scale is greater than 1.0 which introduces fake precision.')

    valid_thresh_modes = ('none', 'manual', 'otsu')
    if thresh_mode.lower() not in ('manual', 'otsu'):
        raise ValueError(f"Invalid thresh_mode. Valid values are: {', '.join(valid_thresh_modes)}.")

    if (thresh_mode.lower() == 'manual') and ((thresh_value >= 1) or (thresh_value <= 0)):
        raise ValueError(f'For manual thresholding, thresh_value must be in the range (0, 1), '
                         f'but it is {thresh_value:.3f}.')

    # Open the image and convert it to grayscale.
    gray_image = imread(fname=image_path, as_gray=True)  # Note gray_image has a dtype of float, i.e. [0, 1].
    if verbose_log:
        logger.debug(f"Opened image '{image_path}'.")

    # Apply the scale.
    gray_image = resize_image(gray_image, scale, verbose_log)

    # The image must be rotated -90 degrees to account for the
    # difference between the XY directions in Abaqus and the picture.
    gray_image = rot90(gray_image, -1)
    if verbose_log:
        logger.debug('Rotated the image -90 degrees to account for '
                     'the difference between the XY directions in Abaqus and the picture.')

    # Denoise the image using a Bilateral filer.
    if denoise:
        gray_image = denoise_bilateral(gray_image)
        if verbose_log:
            logger.debug('Denoised the image using a Bilateral filer.')

    if thresh_mode.lower() == 'none':
        binary_image = gray_image
        if verbose_log:
            logger.debug('No thresholding applied to the image.')
    else:
        # Apply Otsu’s method to make a binary image.
        if thresh_mode == 'manual':
            pass  # thresh_value is assumed to be correct.
        elif thresh_mode == 'otsu':
            thresh_value = threshold_otsu(gray_image)
        else:
            raise ValueError(f"Invalid thresh_mode. Valid values are: {', '.join(valid_thresh_modes)}. "
                             f"This should have been caught earlier.")
        # Threshold the image based on the thresh_value value
        binary_image = gray_image > thresh_value
        if verbose_log:
            logger.debug(f'Applied a {thresh_mode} threshold to the image with a value of {thresh_value:.3f}.')

    # Show the image.
    if show_image:
        fig, axes = plt.subplots(ncols=2, figsize=(9, 4), sharey='all')
        plt.get_current_fig_manager().set_window_title('Image Preview (Close to Continue)')
        ax = axes.ravel()
        ax[0] = plt.subplot(1, 2, 1)
        ax[1] = plt.subplot(1, 2, 2)
        ax[0].imshow(gray_image, cmap=plt.cm.gray)
        ax[0].set_title('Opened Image (Grayscale)')
        ax[0].axis('off')
        ax[1].imshow(binary_image, cmap=ListedColormap(['black', 'white']))
        ax[1].set_title('Binary Image')
        ax[1].axis('off')
        num_dark_px = len(where(binary_image == 0)[0]) / binary_image.size
        num_light_px = len(where(binary_image == 1)[0]) / binary_image.size
        ax[1].annotate(f'Light Pixels = {100 * num_light_px:3.2f}%, Dark Pixels = {100 * num_dark_px:3.2f}%',
                       xy=(0.5, -0.05), xycoords='axes fraction', ha='center')
        plt.show(block=True)

    # Return the binary mask.
    final_log_msg = f"Created a binary mask from the image at '{image_path}' with a scale of {scale:.2f}."
    if final_log:
        logger.info(final_log_msg)
    else:
        if verbose_log:
            logger.debug(final_log_msg)
    return binary_image


def mask_from_image_sequence(load_pattern: str, scale: float = 1.0,
                             denoise: bool = True,
                             thresh_mode: str = 'otsu', thresh_value: float = None):
    """Return a boolean mask by opening and thresholding an image sequence.

    This function opens all images using the :func:`mask_from_image` function,
    applies the scale, and returns the final 3D binary mask.

    This function opens the images with a scale of 1.0 and then resizes them,
    meaning that a lot of RAM may be required.

    Args:
        load_pattern: A pattern describing the path of all images in the sequence.
                      Use the *?* symbol as a placeholder for a single character.
        scale: Scale to be applied to the images. Note that a scale greater than 1.0
               will introduce fake precision by interpolating the data and issues a warning.
        thresh_mode: The kind of threshold to be applied to the image.
                     Valid values are 'none', 'manual', and 'otsu'. Defaults to 'otsu'.
        thresh_value: If *thresh_mode* is set to 'manual', should be a float in the range (0, 1)
                      Which is used as a threshold for binarizing the image.
                      Defaults to *None* which raises an error when used.
        denoise: If set to True, the image will be denoised using a Bilateral filter.

    Returns:
        The binary mask derived from the image sequence.
    """
    if scale <= 0.0:
        raise ValueError('scale must be positive.')
    if scale > 1.0:
        warn('scale is greater than 1.0 which introduces fake precision.')

    valid_thresh_modes = ('none', 'manual', 'otsu')
    if thresh_mode.lower() not in ('manual', 'otsu'):
        raise ValueError(f"Invalid thresh_mode. Valid values are: {', '.join(valid_thresh_modes)}.")

    if (thresh_mode.lower() == 'manual') and ((thresh_value >= 1) or (thresh_value <= 0)):
        raise ValueError(f'For manual thresholding, thresh_value must be in the range (0, 1), '
                         f'but it is {thresh_value:.3f}.')

    # Create an ImageCollection function which loads the images using
    # the mask_from_image() function. No scaling is applied and denoising is done if requested.
    image_coll = ImageCollection(load_pattern=load_pattern, conserve_memory=False,
                                 load_func=mask_from_image,
                                 scale=1.0, denoise=denoise,
                                 thresh_mode=thresh_mode, thresh_value=thresh_value,
                                 show_image=False, verbose_log=False, final_log=False)

    # Make sure the collection has an appropriate number of images.
    if len(image_coll) == 0:
        raise ValueError('The image collection is empty. '
                         'The most likely reason is an incorrect load_pattern.')
    elif len(image_coll) == 1:
        RuntimeWarning('Only one image has been loaded into the image collection. This may indicate misuse '
                       'or an incorrect load_pattern.')
    else:
        logger.info(f'Loaded {len(image_coll)} images in the image collection')

    full_image = moveaxis(image_coll.concatenate(), 0, -1)
    # Apply the scale.
    full_image = resize_image(full_image, scale, verbose_log=True)
    return full_image


def resize_image(image: ndarray, scale: float, verbose_log: bool = True) -> ndarray:
    """Resize an image. This function is not intended for standalone use.

    Args:
        image: The image to be resized.
               If the input image is a boolean mask, nearest-neighbor interpolation
               will be used without antialiasing and the output will be cast to a boolean mask.
               Otherwise, a simple resizing operation with the default parameters
               of ``skimage.transform.rescale()`` will be attempted.
        scale: The scale to be applied. If equal to 1.0, the function logs the call and returns.
        verbose_log: If set to True, logs will be verbose, otherwise no entry will be made.

    Returns:
        The resized image which is of the same *dtype* as the input.
    """
    if scale == 1.0:
        if verbose_log:
            logger.debug('Scale was equal to 1.0. No resizing was performed.')
        return image

    if image.dtype == bool:
        # For boolean images, interpolation order must be 0, which means that
        # the nearest-neighbor interpolation method will be used.
        # Also, antialiasing must be turned off.
        # See https://github.com/scikit-image/scikit-image/issues/4292
        # and https://github.com/scikit-image/scikit-image/issues/4998.
        image = rescale(image, scale, order=0, anti_aliasing=False)

        # Rescale returns a float array which needs to be converted to bool.
        if image.dtype == float:
            if array_equal(unique(image), [0.0, 1.0]):
                image = image.astype(bool)
            else:
                raise RuntimeError('It is expected that skimage.transform.rescale '
                                   'returns a float array with only 0.0 and 1.0 values, '
                                   'which was not the case.')
    else:
        # Image is not a boolean.
        image = rescale(image, scale, anti_aliasing=True)

    if verbose_log:
        logger.debug('Applied a scale of %.2f.', scale)
    return image



"""The voxelpart package contains the main VoxelPart class and its methods.

See the :ref:`voxel-part` section for a complete explanation
of the basic concepts.
"""
import logging
import textwrap
from pathlib import Path
from typing import Union

import numpy as np
from numpy import ndarray, array

from . import __version__, __website__
from .helper import is_name_valid, read_configuration, process_working_dir
from .logger_conf import setup_main_logger
from .mask.function import mask_from_function
from .mask.image import mask_from_image, mask_from_image_sequence
from .mask.random import random_binary_mask
from .mask.shape import Circle, Sphere
from .mask.tpms import tpms_dict
from .output import write_abaqus_inp

logger = logging.getLogger(__name__)


class VoxelPart:
    """A class representing a rectangular or cuboid part made of voxels."""

    def __init__(self, size: tuple[int, int, int] | tuple[int, int],
                 base_material: int = 0,
                 voxel_size: tuple[float, float, float] | tuple[float, float] = (1.0, 1.0, 1.0),
                 dtype: str = 'uint8', name: str = 'unnamed', description: str = '',
                 working_dir: str | Path = None,
                 overwrite_logs: bool = True, log_debug: bool = False,
                 display_log: bool = True, main_logger_exists: bool = False):
        """
        Args:
            size: The tuple *(size_x, size_y, size_z)* which determines
                  the number of voxel elements in the three dimensions.

                  For 2D structures, *size_z* can be omitted,
                  but some parts of the program take it to be 1 for calculations.

                  This parameter determines the shape of the :attr:`data` attribute
                  and therefore must contain integers.

            base_material: The value used for filling :attr:`data` when the object is created.

                           Make sure it is within the range specified by the *dtype* parameter
                           (See the :ref:`materials` section).
                           Defaults to 0 which represents empty space.

            voxel_size: A tuple containing two or three floats which determines the size
                        of a voxel in the three directions.

                        For example, if the tuple (0.02, 0.1, 1.5) is specified,
                        each voxel will have those dimensions in the x, y, and z directions.

                        If a part is 2D, the third value can be omitted and the program assigns 1.0
                        as the rest of the library requires *voxel_size* to have three elements.

                        All values should be between 1E-5 and 1E+3.

            dtype: Data type used for creation of :attr:`data`.
                   Must be an unsigned integer type. Users are advised to study
                   the :ref:`materials` section for a thorough explanation of this parameter.

                   Defaults to *'uint8'ex_c11_* which allows for 256 materials in the model.

            name: Name of the voxel part which is used in a variety of places, including when exporting the part.
                  Must be valid according to the documentation for the :func:`.helper.is_name_valid` function.

                  Defaults to *'unnamed'*.

            description: A short description of the part which is used
                         in a variety of places, including when exporting the part.
                         Note that Abaqus™ only uses the first 80 characters of the string.
                         Defaults to an empty string.

            working_dir: Path to the folder where the final results, temporary file, and log files will be stored.
                         If set to *None* a suitable folder is automatically created in the user's home directory.

            overwrite_logs: If set to True, and the log file already exists, it will be overwritten.
                            Otherwise, the file will be opened in append mode.

            log_debug: If set to True, debug information will be written to program log.

            display_log: If set to True, all logs will be displayed on the screen
                         in addition to being written to file.

            main_logger_exists: If set to True, indicates that the main logger has already been created elsewhere.
                                Defaults to False which creates the main logger.
        """

        # Validate dtype.
        if not dtype.lower() in ('uint8', 'uint16', 'uint32', 'uint64'):
            raise ValueError("dtype can only be one of the following strings: "
                             "'uint8', 'uint16', 'uint32', 'uint64'")
        self.dtype = dtype
        """Data type used for the part."""

        # Set a temporary value for _data which is used by the data property.
        self._data = None

        # It seems that numpy.zeros has a special implementation which
        # makes it faster. numpy.ones is the same as numpy.fill.
        # Source: https://stackoverflow.com/questions/31498784.
        if base_material == 0:
            self.data = np.zeros(shape=size, dtype=dtype.lower())  #: data attribute.
        else:
            self.data = np.full(shape=size, fill_value=base_material, dtype=dtype.lower())
            """The *data* property of the *VoxelPart* instance.
            See the :ref:`data_structure` section for an in depth discussion."""

        # Validate and set voxel_size. Make sure that it has three elements.
        voxel_size = np.array(voxel_size, dtype='float')  # This catches strings and such.
        if voxel_size.shape != (3,):
            if voxel_size.shape == (2,):  # Add 1.0 as the third element.
                voxel_size = np.append(voxel_size, 1.0)
            else:
                raise ValueError('Invalid value for voxel_size.')
        if any(voxel_size <= 1E-5) or any(voxel_size >= 1E+3):
            raise ValueError('Invalid value for voxel_size. Values should be between 1E-5 and 1E+3.')
        self.voxel_size: ndarray = voxel_size
        """A numpy array containing three floats which determines the size of a voxel in the three directions."""

        # Validate name.
        if not is_name_valid(name):
            raise ValueError('Invalid name. Check the documentation for validity criteria.')
        self.name: str = name
        """Name of the voxel part which is used in a variety of places, including when exporting the part."""

        # Validate description.
        if not (isinstance(description, str) and description.isascii() and description.isprintable()):
            raise ValueError('Invalid description.')
        self.description: str = textwrap.fill(description, width=80)
        """A short description of the part which is used in a variety of places, including when exporting the part."""

        # Process and Validate working_dir.
        self.working_dir: Path = process_working_dir(working_dir, name)
        """Path to the folder where the final results, temporary file, and log files will be stored."""

        # Create an empty dictionary for element and node sets.
        self.elem_sets: dict = dict()
        """Dictionary in which keys are the names of the element sets
        and the values are and IDs of the elements in that set."""
        self.node_sets: dict = dict()
        """Dictionary in which keys are the names of the node sets
        and the values are and IDs of the elements in that set."""

        # Create variables for bcs and their sets.
        self._bc_type = None
        self._bc_nodeset_explicit = False
        self._bc_nodeset_simple = False
        self._dummy_node_dict = dict()

        # Create and configure the logger.
        if not main_logger_exists:
            self._log_file_path = setup_main_logger(part_name=name, working_dir=working_dir,
                                                    display_log=display_log, overwrite_logs=overwrite_logs,
                                                    log_debug=log_debug)
        else:
            self._log_file_path = None

        # Log creation of the object.
        logger.info('\n** Created using VCAMS v%s.'
                    '\n** VCAMS is a free and open source program available at: %s'
                    '\n** Author: Mohammadreza Khoshbin (www.mkhoshbin.com)\n',
                    __version__, __website__)
        logger.info("A VoxelPart object named '%s' was created" +
                    " with %s elements and an initial element value of %u.",
                    name, '*'.join(str(s) for s in size), base_material)

    @property
    def instance_name(self) -> str:
        """Name of the part instance which is name of the part + '-Ins'. Used for output to Abaqus™ input file."""
        return self.name + '-Ins'

    @property
    def size(self) -> ndarray:
        """Size of the *VoxelPart* instance.

        This is the shape of the part's *data* property, returned as a NumPy array.
        If the instance is 2D, it will have two elements, otherwise it will have three."""
        return array(self.data.shape)

    @property
    def real_size(self) -> ndarray:
        """Real size of the *VoxelPart* instance.

        This is defined as ``part.size * voxel_size``, and is returned as a NumPy array.
        If the instance is 2D, it will have two elements, otherwise it will have three."""
        return np.array([self.size[i] * self.voxel_size[i] for i in range(len(self.size))])

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value: ndarray):
        if not isinstance(value, ndarray):
            raise ValueError('data must be a numpy ndarray.')
        if not value.flags.c_contiguous:
            raise ValueError('data must be C-continuous.')
        if (self.data is not None) and (self.data.ndim != value.ndim):
            # Because in __init__, data is initially None.
            raise ValueError('The new value must have the exact ndim as the current data attribute.')
        if value.dtype != self.dtype:
            raise ValueError('The new value must have the exact dtype as the VoxelPart instance.')
        self._data = value.astype(dtype=self.dtype, order='C', casting='safe', subok=True, copy=True)

    def __del__(self):
        """Delete the object. The respective loggers are also flushed and closed."""
        logger_list = (logging.getLogger(__name__),)
        for lg in logger_list:
            for h in lg.handlers:
                # Handlers are only flushed and closed,
                # but not removed because I don't think it's necessary.
                h.flush()
                h.close()

    def __len__(self):
        return np.prod(self.size)

    def output_abaqus_inp(self, file_name: str, elem_code: str, dim: str,
                          material_elem_sets: str | tuple, custom_elem_sets: bool = True,
                          keep_temp_files: bool = False) -> Path:
        """Output the part to an Abaqus™ input file.

        Only the elements selected by the *material_elem_sets* parameter are selected,
        and afterwards they are grouped into sets by the material code.
        If *custom_elem_sets* is True, the custom element set are also included.
        If an element is part of a custom element set but is not part of the selected materials,
        It is not written to the output.

        This function simply calls :func:`.output.write_abaqus_inp`, except for the
        *scale* and *folder_path* parameters which are equal to :attr:`.VoxelPart.voxel_size`
        and :attr:`.VoxelPart.working_dir`, respectively.

        Args:
            file_name: Name of the file. Must be valid according to the documentation
                       for the :func:`.helper.is_name_valid` function and should not contain file extensions.
            elem_code: An uppercase string denoting the element code assigned to *all* elements in the model.
                       It must be a valid Abaqus element code such as *'CPE4R'* or *'C3D8R'*.
                       This parameter is not validated so care should be taken regarding validity and compatibility.
                       Currently, only 2D and 3D linear elements are supported.
                       To get around this, you can convert to quadratic elements after importing the model to Abaqus.
            dim: Dimensionality of the output part. Valid values are *'2D'* and *'3D'*.
            material_elem_sets: One of the following:

                                  + *'All'* which outputs all materials in the VoxelPart.
                                  + *'Non-Empty'* which outputs all non-zero (=non-empty) materials in the VoxelPart.
                                  + A tuple of integer material codes corresponding
                                    to the materials that should be written to the output.

            custom_elem_sets: If set to True, custom element sets will be written to the output.
            keep_temp_files: If set to True, temporary files will not be deleted. Used for debugging.

        Returns:
            Path object pointing to final Abaqus™ input file.
        """
        # Logging is done by the called function.
        return write_abaqus_inp(self, file_name=file_name,
                                elem_code=elem_code, dim=dim,
                                scale=tuple(self.voxel_size),
                                material_elem_sets=material_elem_sets,
                                custom_elem_sets=custom_elem_sets,
                                keep_temp_files=keep_temp_files)

    def add_custom_elem_set(self, name: str, ids: tuple | ndarray, replace: bool = True):
        """Add a custom element set to the part.

        Args:
            name: Name of the set. Must be valid according to the documentation for :func:`.helper.is_name_valid`.
            ids: A tuple or numpy array of integer element IDs to be added to the set.
                 The IDs should start at zero (zero-based indexing), and the proper value is output later.
                 The method converts it into a sorted, unique numpy array, but no other validation is performed.
            replace: If set to True and a set with the same name already  exists, the new set replaces the old one.
                     Otherwise, an error is raised.

                     Defaults to True.
        """
        if not is_name_valid(name):
            raise ValueError('Invalid name. Check the documentation for validity criteria.')
        if name in self.elem_sets and not replace:
            raise RuntimeError("An element set with the name '%s' already exists." % name)

        self.elem_sets[name] = np.unique(ids).astype('uint32')
        logger.debug("Added custom element set '%s' with %u elements.", name, len(self.elem_sets[name]))

    def add_node_set(self, name: str, ids: tuple | ndarray, replace: bool = True):
        """Add a node set to the part.

        Args:
            name: Name of the set. Must be valid according to the documentation for :func:`.helper.is_name_valid`.
            ids: A tuple or numpy array of integer node IDs to be added to the set.
                 The IDs should start at zero (zero-based indexing), and the proper value is output later.
                 The method converts it into a sorted, unique numpy array, but no other validation is performed.
            replace: If set to True and a set with the same name already  exists, the new set replaces the old one.
                     Otherwise, an error is raised.

                     Defaults to True.
        """
        if not is_name_valid(name):
            raise ValueError('Invalid name. Check the documentation for validity criteria.')
        if name in self.node_sets and not replace:
            raise RuntimeError(f"A node set with the name '{name}' already exists.")
        self.node_sets[name] = np.unique(ids).astype('uint32')
        logger.debug(f"Added custom node set '{name}' with {len(self.node_sets[name])} elements.")

    def add_bc(self, bc_type: str | None = None,
               explicit_nodesets: bool = False, simple_nodesets: bool = False):
        """Define a boundary condition (BC) for the part.
        Refer to the :ref:`boundary-conditions` section for a full explanation of the available BCs.

        The part must be a full square or cuboid. Also, if a BC is requested using *bc_type*,
        all necessary node sets are also created.
        For most use cases, that is the only parameter that must be specified.

        Args:
            bc_type: Type of the BC to be defined. Valid values are:

                     + None: No BCs will be defined.
                     + 'Nodeset Only': Only the node sets will be defined according to the other parameters.
                     + 'Linear Displacement': :ref:`boundary-conditions-lin-disp` will be created.
                     + 'Periodic': :ref:`boundary-conditions-pbc` will be created.

            explicit_nodesets: If True, explicit node sets are created for vertices, edges, and faces.
                               as described in the section titled :ref:`boundary-conditions-pbc`.
                               Defaults to *False*.
            simple_nodesets: If True, simplified node sets are created
                             which are the full faces for the 3D models or the edges for the 2D models,
                             as described in the section titled :ref:`boundary-conditions-lin-disp`.
                             Defaults to *True*.
        """
        if bc_type is None:
            self._bc_type = None
        elif bc_type.upper() in ['NODESET ONLY', 'LINEAR DISPLACEMENT', 'PERIODIC']:
            self._bc_type = bc_type.upper()
        else:
            raise ValueError('Invalid value for bc_type.')

        if (bc_type.upper() == 'NODESET ONLY') and not any([explicit_nodesets, simple_nodesets]):
            raise ValueError("bc_type is set to 'NODESET ONLY', but no node sets are requested."
                             "At least one of explicit_nodeset and simple_nodeset must be set to True.")
        self._bc_nodeset_explicit = explicit_nodesets
        self._bc_nodeset_simple = simple_nodesets

    def apply_mask(self, mask: ndarray, value: int):
        """Use a Boolean mask to select some elements of the part's :attr:`data` array
        and change them to a *value*.

        Args:
            mask: The Boolean mask to be used.
            value: Integer value to be assigned to the elements of the :attr:`data` attribute
                   where the Boolean mask is True.
        """
        # Make sure mask is a Boolean mask.
        if not mask.dtype == bool:
            raise ValueError("mask.dtype is not 'bool'.")
        # Make sure mask and self.data have the same shape.
        if not np.array_equal(self.size, mask.shape):
            if self.data.ndim == 2 and mask.shape[2] == 1:
                pass
            else:
                raise ValueError('mask is not of the same shape as VoxelPart.data.')
        # Make sure mask and self.data have the same order (Fortran or C contiguity).
        if mask.flags.f_contiguous != self.data.flags.f_contiguous:
            raise ValueError('mask is not of the same order (Fortran or C contiguity) as VoxelPart.data.')
        # Make sure value is a nonzero integer within the bounds of self.data.dtype.
        if not float(value).is_integer():
            raise ValueError('value is not an integer.')
        if value < 0:
            raise ValueError('value is less than zero.')
        if value > np.iinfo(self.data.dtype).max:
            raise ValueError(f'value is larger than the maximum supported by self.data.dtype, '
                             f'which is {np.iinfo(self.data.dtype).max}.')
        # Apply the mask to self.data.
        np.putmask(self.data, mask, value)

    def _add_dummy_nodes(self, fixed: bool = False, single_node: bool = False, three_nodes: bool = False):
        """Add the dummy nodes to the part.
        See :ref:`the relevant section on BCs <boundary-conditions-nodeset_only>` for more information.

        Args:
            fixed: If set to True, the dummy node for the fixed point is added with a node ID of 999999999.
            single_node: If set to True, the dummy node for a single moving point is added
                         with a node ID of 999999998.
            three_nodes: If set to True, the dummy node for three moving points is added
                         with a node IDs of 999999996, 999999997, and 999999998.
        """
        if not single_node ^ three_nodes:
            raise ValueError("Exactly one of single_node or three_nodes must be True.")
        if fixed:
            self._dummy_node_dict['RP0-NodeSet'] = 999999999
        if single_node:
            self._dummy_node_dict['RP1-NodeSet'] = 999999998
        if three_nodes:
            self._dummy_node_dict['RP1-NodeSet'] = 999999996
            self._dummy_node_dict['RP2-NodeSet'] = 999999997
            self._dummy_node_dict['RP3-NodeSet'] = 999999998

    def _return_material_elem_set(self, mat_code: int, num_padding: int = 0) -> tuple[str, ndarray]:
        """Return the IDs of the elements in the part that correspond to the given material code
        and a suitable name for the set.

        The name is a string with 'MAT-' prepended to the material code.

        Args:
            mat_code: Integer specifying the material for which element IDs must be found.
            num_padding: Number of padding zeros for the numerical portion of the material name.
                         Defaults to 0, which means no padding.

        Returns:
            A tuple where the first element is the material name,
            and the second element is a 1-D numpy array of element IDs.
        """
        name = 'MAT-{mat_code:0{num_padding}d}'.format(mat_code=mat_code, num_padding=num_padding)
        # Source: https://stackoverflow.com/a/32413139/7180705
        elem_ids = np.ravel_multi_index(multi_index=np.nonzero(self.data == mat_code),
                                        dims=self.size, mode='raise', order='C').astype('uint32')
        return name, elem_ids


def voxelpart_from_image(image_dim: str, image_path: str,
                         scale: float = 1.0, denoise: bool = True,
                         show_image: bool = False,
                         thresh_mode: str = 'otsu', thresh_value: float = None,
                         background_material: int = 0, foreground_material: int = 1,
                         voxel_size: tuple[float, float, float] | tuple[float, float] = (1.0, 1.0, 1.0),
                         dtype: str = 'uint8', name: str = 'unnamed',
                         description: str = '', working_dir: str | Path = None,
                         overwrite_logs: bool = True, log_debug: bool = False, display_log: bool = True,
                         **kwargs):
    """Create a :class:`VoxelPart` instance from a 2D or 3D image.

    This function creates a Boolean mask using either
    :func:`~vcams.mask.image.mask_from_image` or :func:`~vcams.mask.image.mask_from_image_sequence`,
    and then creates a :class:`VoxelPart` with the same size as the mask
    and properties given as input.
    The mask is then applied to the part using the *VoxelPart*'s :meth:`~VoxelPart.apply_mask` method.

    Args:
        image_dim: Dimensionality of the input image. Valid values are '2D' and '3D'.
        image_path:

          - For 2D images: The path string referring to the 2D image.
          - For 3D images: A string containing a loading pattern
            which describes the path of all images in the sequence.
            Use the *?* symbol as a placeholder for a single character.

        show_image: If set to True, the opened image and the final binary image
                    will be shown side by side in a figure.
                    The program may be paused while the window is open.
                    Currently, this has only been implemented for 2D images
                    and will be silently ignored for 3D images.
        scale: Scale to be applied to the image(s). Note that a scale greater than 1.0
               will introduce fake precision by interpolating the data and issues a warning.
        denoise: If set to True, the image will be denoised using a Bilateral filter.
        thresh_mode: The kind of threshold to be applied to the image.
                     Valid values are 'none', 'manual', and 'otsu'. Defaults to 'otsu'.
        thresh_value: If *thresh_mode* is set to 'manual', should be a float in the range (0, 1)
                      Which is used as a threshold for binarizing the image.
                      Defaults to *None* which raises an error when used.
        background_material: After the image is binarized, this material will be assigned to the *OFF* pixels.
        foreground_material: After the image is binarized, this material will be assigned to the *ON* pixels.
        voxel_size: See the documentation for the :meth:`VoxelPart.__init__` method.
        dtype: See the documentation for the :meth:`VoxelPart.__init__` method.
        name: See the documentation for the :meth:`VoxelPart.__init__` method.
        description: See the documentation for the :meth:`VoxelPart.__init__` method.
        working_dir: See the documentation for the :meth:`VoxelPart.__init__` method.
        overwrite_logs: See the documentation for the :meth:`VoxelPart.__init__` method.
        log_debug: See the documentation for the :meth:`VoxelPart.__init__` method.
        display_log: See the documentation for the :meth:`VoxelPart.__init__` method.

    Note that the size parameter used in the :meth:`VoxelPart.__init__` method
    is not one of the inputs and is determined from the image.
    """

    part_log_file_path = setup_main_logger(part_name=name, working_dir=working_dir,
                                           display_log=display_log, overwrite_logs=overwrite_logs,
                                           log_debug=log_debug)
    logger.info('Attempting to create a VoxelPart instance from image(s).')

    if image_dim.upper() not in ['2D', '3D']:
        raise ValueError("image_dim can only be one of '2D' or '3D'.")
    if image_dim.upper() == '2D':
        image_mask = mask_from_image(image_path=image_path, scale=scale,
                                     denoise=denoise, show_image=show_image,
                                     thresh_mode=thresh_mode, thresh_value=thresh_value)
    else:  # dim.upper() == '3D'
        image_mask = mask_from_image_sequence(load_pattern=image_path,
                                              scale=scale, denoise=denoise,
                                              thresh_mode=thresh_mode, thresh_value=thresh_value)
    # Create a VoxelPart instance.
    part_shape = image_mask.shape
    part = VoxelPart(size=part_shape, base_material=background_material, voxel_size=voxel_size,
                     dtype=dtype, name=name, description=description, working_dir=working_dir,
                     overwrite_logs=overwrite_logs, log_debug=log_debug, display_log=display_log,
                     main_logger_exists=True)
    part._log_file_path = part_log_file_path
    part.apply_mask(mask=image_mask, value=foreground_material)
    return part


def from_config_file(file_path: Union[str, Path]) -> VoxelPart:
    """Create a :class:`VoxelPart` object from a configuration file.

    Args:
        file_path: Full path to the configuration file. This file is usually created using the GUI
                   and although you can create or edit one, it's not recommended.
                   Scripts are much easier to work with and
                   this function is meant only as a bridge between the library and its GUI.

    Returns:
        The :class:`VoxelPart` object created based on the configuration file.
    """
    (part_creation_dict, part_manipulation_dict, bc_dict, output_dict) = read_configuration(file_path)

    logger.info('The model is being created from a configuration file loaded from %s' % file_path)
    modeling_mode = part_manipulation_dict['modeling_mode']
    if modeling_mode not in ('4', '5'):
        part = VoxelPart(**part_creation_dict)
    if modeling_mode == '0':  # No action selected. This is technically invalid, but consider it to be '1'.
        pass
    if modeling_mode == '1':  # No Further Manipulation.
        pass
    elif modeling_mode == '2':  # Random Element Dispersion
        boolean_mask = random_binary_mask(part=part, true_fraction=part_manipulation_dict['random_phase_fraction'])
        part.apply_mask(mask=boolean_mask, value=part_manipulation_dict['random_phase_matcode'])
    elif modeling_mode == '3':  # TPMS
        boolean_mask = mask_from_function(mask_shape=part.size,
                                          func=tpms_dict[part_manipulation_dict['tpms_type']],
                                          part=part,
                                          l=part_manipulation_dict['tpms_length'],
                                          c=part_manipulation_dict['tpms_constant'])
        part.apply_mask(mask=boolean_mask, value=part_manipulation_dict['tpms_fill_value'])
    elif modeling_mode == '4':  # Image Processing (Single 2D Image)
        part = voxelpart_from_image(image_dim='2D',
                                    image_path=part_manipulation_dict['single_image_path'],
                                    scale=part_manipulation_dict['single_image_scale'],
                                    thresh_mode=part_manipulation_dict['single_image_thresh_mode'],
                                    thresh_value=part_manipulation_dict['single_image_thresh_value'],
                                    denoise=part_manipulation_dict['single_image_denoise'],
                                    background_material=part_creation_dict['base_material'],
                                    foreground_material=2,  # This is hardcoded.
                                    **part_creation_dict)
    elif modeling_mode == '5':  # Stack of 2D images for a 3D part.
        part = voxelpart_from_image(image_dim='3D',
                                    image_path=part_manipulation_dict['multi_image_path'],
                                    scale=part_manipulation_dict['multi_image_scale'],
                                    thresh_mode=part_manipulation_dict['multi_image_thresh_mode'],
                                    thresh_value=part_manipulation_dict['multi_image_thresh_value'],
                                    denoise=part_manipulation_dict['multi_image_denoise'],
                                    background_material=part_creation_dict['base_material'],
                                    foreground_material=2,  # This is hardcoded.
                                    **part_creation_dict)
    elif modeling_mode == '6':  # Planar Composite (Circular Inclusions)
        for row in part_manipulation_dict['circle_list']:
            circle_obj = Circle(id=0, xc=float(row[0]), yc=float(row[1]), r=float(row[2]))
            part.apply_mask(mask=circle_obj.calculate_mask(part_shape=part.size, voxel_size=part.voxel_size),
                            value=int(row[3]))
    elif modeling_mode == '7':  # Spatial Composite (Spherical Inclusions)
        for row in part_manipulation_dict['sphere_list']:
            circle_obj = Sphere(id=0, xc=float(row[0]), yc=float(row[1]), zc=float(row[2]), r=float(row[3]))
            part.apply_mask(mask=circle_obj.calculate_mask(part_shape=part.size, voxel_size=part.voxel_size),
                            value=int(row[4]))
    else:
        raise ValueError("Invalid value '%s' for part_manipulation_dict['modeling_mode']." % modeling_mode)

    bc_type = bc_dict['bc_type']
    if bc_type == '0':  # No BC.
        pass
    elif bc_type == '1':  # Sets only.
        part.add_bc(bc_type='NODESET ONLY', explicit_nodesets=True, simple_nodesets=True)
    elif bc_type == '2':  # Linear Displacement Boundary Conditions.
        part.add_bc(bc_type='Linear Displacement')
    elif bc_type == '3':  # Periodic Boundary Condition.
        part.add_bc(bc_type='Periodic')
    else:
        raise ValueError("Invalid value '%s' for bc_dict['bc_type']" % bc_type)

    part.output_abaqus_inp(**output_dict)

    logger.info('Creation of the model from the configuration file completed successfully.')
    return part



"""Classes defining Triply Periodic Minimal Surfaces (TPMS)
which can be used to create Boolean masks.

The resulting masks can then be used
for manipulating a :class:`~vcams.voxelpart.VoxelPart` instance
using its :meth:`~vcams.voxelpart.VoxelPart.apply_mask` method.
See the :ref:`predefined-tpms` section for a complete explanation
of the basic concepts.
"""
from abc import ABC, abstractmethod
from typing import Callable

from numpy import cos, pi, sin, ndarray


class BaseTpms(ABC):
    """Abstract base class describing a Triply Periodic Minimal Surface (TPMS).
    Note that the subclasses will be static.

    All TPMS classes must inherit from this class.
    Subclasses define the level set function *func* describing the surface in 3D space,
    and the three attributes *name*, *tpms_id*, and *formula* which are used by PyQt
    for the GUI.
    """
    @property
    @abstractmethod
    def name(self):
        """Name of the TPMS which is used in the GUI."""
        pass

    @property
    @abstractmethod
    def tpms_id(self):
        """An integer id that is unique to this TPMS class and is used in the GUI."""
        pass

    @property
    @abstractmethod
    def formula(self):
        """Formula of the TPMS which must be an inline latex math
        compatible with PyQt and is used in the GUI."""
        pass

    @staticmethod
    @abstractmethod
    def func(*args):
        """The level set function *func* describing the surface in 3D space.
        It must be compatible with :func:`vcams.mask.function.mask_from_function`.
        """
        pass


class TpmsSchwarzP(BaseTpms):
    """Static class describing a Schwarz Primitive (P) triply periodic minimal surface
    with the following equation:

    .. math::
       \\Phi = cos(\\frac{2\\pi}{l}x) + cos(\\frac{2\\pi}{l}y) + cos(\\frac{2\\pi}{l}z) - c = 0
    """
    @staticmethod
    def func(x: ndarray, y: ndarray, z: ndarray, l: float, c: float | Callable) -> float | ndarray:  # noqa: E741
        """Function describing a Schwarz Primitive (P) triply periodic minimal surface.

        Args:
            x: A numpy 1D array of x-coordinates.
            y: A numpy 1D array of y-coordinates.
            z: A numpy 1D array of z-coordinates.
            l: Length of the unit cell in all directions.
            c: Parameter C in the equation.
               It can either be a constant *float*, or a callable
               that's a function of x, y, and z, is vectorized, and returns a float.
               Note that c(x, y, z) is not validated so great care
               should be taken when constructing and passing it.

        Returns:
            An array of floats which may be negative, zero, or positive.
            If scalar values are passed, a float is returned instead of an array.
            See :ref:`level-set-functions` for interpretation of the results.
        """
        p = 2 * pi / l  # Period.
        if callable(c):
            c = c(x, y, z)
        return cos(p * x) + cos(p * y) + cos(p * z) - c

    tpms_id: int = 0
    name: str = 'Schwarz Primitive (P)'
    formula: str = r'$\Phi = cos(\frac{2\pi}{l} x) + cos(\frac{2\pi}{l} y) + cos(\frac{2\pi}{l} z) - c$'


class TpmsSchwarzD(BaseTpms):
    """Static class describing a Schwarz Diamond (D) triply periodic minimal surface
    with the following equation:

    .. math::
       \\Phi = \\enspace &sin(\\frac{2\\pi}{l}x) \\ sin(\\frac{2\\pi}{l}y) \\ sin(\\frac{2\\pi}{l}z)\\\\
                     +\\ &sin(\\frac{2\\pi}{l}x) \\ cos(\\frac{2\\pi}{l}y) \\ cos(\\frac{2\\pi}{l}z)\\\\
                     +\\ &cos(\\frac{2\\pi}{l}x) \\ sin(\\frac{2\\pi}{l}y) \\ cos(\\frac{2\\pi}{l}z)\\\\
                     +\\ &cos(\\frac{2\\pi}{l}x) \\ cos(\\frac{2\\pi}{l}y) \\ sin(\\frac{2\\pi}{l}z) - c = 0
    """
    @staticmethod
    def func(x: ndarray, y: ndarray, z: ndarray, l: float, c: float | Callable) -> float | ndarray:  # noqa: E741
        """Function describing a Schwarz Diamond (D) triply periodic minimal surface.

        Args:
            x: A numpy 1D array of x-coordinates.
            y: A numpy 1D array of y-coordinates.
            z: A numpy 1D array of z-coordinates.
            l: Length of the unit cell in all directions.
            c: Parameter C in the equation.
               It can either be a constant *float*, or a callable
               that's a function of x, y, and z, is vectorized, and returns a float.
               Note that c(x, y, z) is not validated so great care
               should be taken when constructing and passing it.

        Returns:
            An array of floats which may be negative, zero, or positive.
            If scalar values are passed, a float is returned instead of an array.
            See :ref:`level-set-functions` for interpretation of the results.
        """
        p = 2 * pi / l  # Period.
        if callable(c):
            c = c(x, y, z)
        return (sin(p * x) * sin(p * y) * sin(p * z) +
                sin(p * x) * cos(p * y) * cos(p * z) +
                cos(p * x) * sin(p * y) * cos(p * z) +
                cos(p * x) * cos(p * y) * sin(p * z) - c)

    tpms_id: int = 1
    name: str = 'Schwarz Diamond (D)'
    formula: str = (r'$\Phi = sin(\frac{2\pi}{l} x) sin(\frac{2\pi}{l} y) sin(\frac{2\pi}{l} z)$' '\n'
                    r'$+ sin(\frac{2\pi}{l} x) cos(\frac{2\pi}{l} y) cos(\frac{2\pi}{l} z)$' '\n'
                    r'$+ cos(\frac{2\pi}{l} x) sin(\frac{2\pi}{l} y) cos(\frac{2\pi}{l} z)$' '\n'
                    r'$+ cos(\frac{2\pi}{l} x) cos(\frac{2\pi}{l} y) sin(\frac{2\pi}{l} z) - c$')


class TpmsSchwarzG(BaseTpms):
    """Static class describing a Schwarz Gyroid (G) triply periodic minimal surface
    with the following equation:

    .. math::
       \\Phi = \\enspace &sin(\\frac{2\\pi}{l}x) \\ cos(\\frac{2\\pi}{l}y)\\\\
                     +\\ &sin(\\frac{2\\pi}{l}y) \\ cos(\\frac{2\\pi}{l}z)\\\\
                     +\\ &sin(\\frac{2\\pi}{l}z) \\ cos(\\frac{2\\pi}{l}x) - c = 0
    """
    @staticmethod
    def func(x: ndarray, y: ndarray, z: ndarray, l: float, c: float | Callable) -> float | ndarray:  # noqa: E741
        """Function describing a Schwarz Gyroid (G) triply periodic minimal surface.

        Args:
            x: A numpy 1D array of x-coordinates.
            y: A numpy 1D array of y-coordinates.
            z: A numpy 1D array of z-coordinates.
            l: Length of the unit cell in all directions.
            c: Parameter C in the equation.
               It can either be a constant *float*, or a callable
               that's a function of x, y, and z, is vectorized, and returns a float.
               Note that c(x, y, z) is not validated so great care
               should be taken when constructing and passing it.

        Returns:
            An array of floats which may be negative, zero, or positive.
            If scalar values are passed, a float is returned instead of an array.
            See :ref:`level-set-functions` for interpretation of the results.
        """
        p = 2 * pi / l  # Period.
        if callable(c):
            c = c(x, y, z)
        return (sin(p * x) * cos(p * y) +
                sin(p * y) * cos(p * z) +
                sin(p * z) * cos(p * x) - c)

    tpms_id: int = 2
    name: str = 'Schwarz Gyroid (G)'
    formula: str = (r'$\Phi = sin(\frac{2\pi}{l} x) cos(\frac{2\pi}{l} y)$' '\n'
                    r'$+ sin(\frac{2\pi}{l} y) cos(\frac{2\pi}{l} z)$' '\n'
                    r'$+ sin(\frac{2\pi}{l} z) cos(\frac{2\pi}{l} x) - c$')


# Construct a dictionary of tpms ids and their respective classes.
tpms_dict: dict = dict()
"""A dictionary mapping tpms ids to their respective classes, used for the GUI."""
for cls in (TpmsSchwarzP, TpmsSchwarzD, TpmsSchwarzG):
    if cls.tpms_id in tpms_dict.keys():
        raise RuntimeError(('Class %s has a non-unique tpms_id. ' % cls.__name__) +
                           'All ids for the TPMS classes must be rechecked. If you are not the '
                           'developer, contact him. If you have added a new TPMS class, '
                           'check your tpms_id.')
    tpms_dict[cls.tpms_id] = cls



"""Functions used for creating a Boolean mask from a level set function.

The resulting masks can then be used
for manipulating a :class:`~vcams.voxelpart.VoxelPart` instance
using its :meth:`~vcams.voxelpart.VoxelPart.apply_mask` method.
See the :ref:`modeling-techniques` section for a complete explanation
of the basic concepts.
"""
from inspect import isclass
from logging import getLogger
from time import perf_counter
from typing import Callable

import numpy as np
from numpy import any, ceil, logical_or, max, ndarray, squeeze, array, zeros_like

logger = getLogger(__name__)


def mask_from_function(func: Callable | object, wrap_mask: bool = False,
                       do_log: bool = True,
                       part=None,
                       mask_shape: tuple[int, int, int] = None,
                       voxel_size: tuple[float, float, float] = None, **kwargs) -> ndarray:
    """Create a boolean mask based on a function describing a surface.

    Args:
        part (VoxelPart | None): The :class:`~.voxelpart.VoxelPart` based on which the mask is created.
                                 If *None*, arguments *mask_shape* and *voxel_size* must be specified
                                 otherwise they are ignored. Defaults to *None*.
        func: One of the following:

              + A function which evaluates a point and returns a value.
                This function must accept x, y, and z parameters (if not use them)
                and can receive other keyword arguments through *\\**kwargs*.
              + A class that defined a method named *func* working according
                to the aforementioned specs.
                For example, it is possible to pass a subclass of :class:`~.BaseTpms`
                and its *func* method is used for the operation.

              This function can represent anything, for example:

              - A level-set function such as the Schwarz P
                triply periodic minimal surface (TPMS).
              - The equation for a circle. In this case,
                the function receives the z variable, but does not use it.

        wrap_mask: If set to True, the function is wrapped around the boundaries of the working space.
                   This is useful for periodic structures, but is computationally expensive.
        do_log: If set to True, name of the function and elapsed time is
                written to the log at the end of the operation.
        mask_shape: A tuple containing three integers which determine
                    the shape of the returned boolean mask. Ignored if *part* is passed.
        voxel_size: A tuple containing three floats which determine the size of a voxel
                    in the x, y, and z directions. Ignored if *part* is passed.
        **kwargs: Any keyword arguments passed to this function are passed to *func*.
                  If any of them is a vector, care should be taken to ensure that the
                  function can accept them as vectors.

    Returns:
        A numpy ndarray with a dtype of bool representing the resulting boolean mask.
        It can be combined with other masks or be applied using :meth:`.VoxelPart.apply_mask`.
    """

    if part:
        mask_shape = part.size
        voxel_size = part.voxel_size

    # Validate voxel_size and convert to numpy array.
    voxel_size = np.array(voxel_size, dtype=float)
    if len(voxel_size) == 3:
        pass
    elif len(voxel_size) == 2:
        voxel_size = np.append(voxel_size, 1)
    else:
        raise ValueError('voxel_size must have a length of 2 or 3.')

    # Validate mask_shape and convert to numpy array.
    mask_shape = np.array(mask_shape, dtype=float)
    if len(mask_shape) == 3:
        pass
    elif len(mask_shape) == 2:
        mask_shape = np.append(mask_shape, 1)
    else:
        raise ValueError('mask_shape must have a length of 2 or 3.')

    if isclass(func) and hasattr(func, 'func') and callable(func.func):
        func = func.func

    start_time = perf_counter()
    x, y, z = np.ogrid[0:mask_shape[0], 0:mask_shape[1], 0:mask_shape[2]]
    mask = is_voxel_inside(x, y, z, voxel_size, func, **kwargs)
    if wrap_mask:
        # noinspection PyTypeChecker
        wrapping_mask = _find_wrapping_mask(mask, func, mask_shape, voxel_size, **kwargs)
        mask = logical_or(mask, wrapping_mask)
    # This is the regular for loop that can replace the above vectorized statements.
    # It used to be an option, but seeing that it will not work
    # for a wrapping mask it has been removed.
    # mask = np.zeros(mask_shape, dtype=bool)
    # for i in np.arange(mask_shape[0]):
    #     for j in np.arange(mask_shape[1]):
    #         for k in np.arange(mask_shape[2]):
    #             mask[i, j, k] = is_voxel_inside(i, j, k, voxel_size, func, **kwargs)

    if do_log:
        logger.info("Created a mask from the function named '%s' in %.2f seconds.",
                    func.__name__, perf_counter() - start_time)
    return squeeze(mask)


def is_voxel_inside(x: float | ndarray, y: float | ndarray, z: float | ndarray,
                    voxel_size: tuple[float, float, float] | ndarray, func: Callable, **kwargs) -> bool:
    """Determine if a voxel is inside a surface.

    A voxel is always cubic, and 27 points of interest (PoI) are defined on it as shown below:

    .. figure:: /images/voxel-poi.png
       :name: voxel-poi
       :scale: 40%
       :align: center
       :alt: Illustration of the 27 points of interest (PoI) checked in a voxel.

    This function compiles three vectors that together form the coordinates
    of these points and passes them to the vectorized function.
    Any other arguments that the surface function needs are passed to it using \\**kwargs.

    The result is a vector of values. If these values are negative,
    that particular PoI is considered to be inside the surface.
    If more than half of the 27 PoIs are inside,
    the voxel is considered inside the surface defined by *func*.
    This function is also vectorized for use with the output of numpy.ogrid,
    which is faster but that may consume a lot of memory.

    Note that this function performs no input validations to ensure efficiency.
    Extreme caution must be used when calling it to ensure correct results.

    Args:
        x: The x index of the voxel. It is multiplied by *voxel_size[0]* to obtain the x coordinate.
        y: The y index of the voxel. It is multiplied by *voxel_size[1]* to obtain the y coordinate.
        z: The z index of the voxel. It is multiplied by *voxel_size[2]* to obtain the z coordinate.
        voxel_size: A tuple containing exactly three floats
                    which determine the size of a voxel in the x, y, and z directions.
                    For 2D parts, the third dimension can be any float, preferably 1.0.
        func: See the *func* parameter of :func:`mask_from_function`.
        **kwargs: Any keyword arguments passed to this function are passed to *func*.
                  If any of them is a vector, care should be taken to ensure that the
                  function can accept them as vectors.

    Returns:
        Returns True if the voxel is inside the surface defined by *func* and False if outside.
    """
    # Contributors are discouraged from modifying this function.

    # The coordinates of the 27 PoI have three unique values along each axis, Which are:
    xx = np.array((x, x + 0.5, x + 1.0)) * voxel_size[0]
    yy = np.array((y, y + 0.5, y + 1.0)) * voxel_size[1]
    zz = np.array((z, z + 0.5, z + 1.0)) * voxel_size[2]

    # These can be combined into a list of coordinates using a number of
    # different methods, including np.meshgrid.
    # But a hard coded approach is faster.
    x_array = np.array((xx[0], xx[0], xx[0], xx[0], xx[0], xx[0], xx[0], xx[0], xx[0],
                        xx[1], xx[1], xx[1], xx[1], xx[1], xx[1], xx[1], xx[1], xx[1],
                        xx[2], xx[2], xx[2], xx[2], xx[2], xx[2], xx[2], xx[2], xx[2]))
    y_array = np.array((yy[0], yy[0], yy[0], yy[1], yy[1], yy[1], yy[2], yy[2], yy[2],
                        yy[0], yy[0], yy[0], yy[1], yy[1], yy[1], yy[2], yy[2], yy[2],
                        yy[0], yy[0], yy[0], yy[1], yy[1], yy[1], yy[2], yy[2], yy[2]))
    z_array = np.array((zz[0], zz[1], zz[2], zz[0], zz[1], zz[2], zz[0], zz[1], zz[2],
                        zz[0], zz[1], zz[2], zz[0], zz[1], zz[2], zz[0], zz[1], zz[2],
                        zz[0], zz[1], zz[2], zz[0], zz[1], zz[2], zz[0], zz[1], zz[2]))

    # Call the function using the coordinate arrays and kwargs.
    # The result is a vector of values. If these values are negative,
    # that particular PoI is considered inside the surface.
    # If more than half of the 27 PoIs are inside,
    # the voxel is considered inside the surface defined by the function.
    return np.count_nonzero(func(x_array, y_array, z_array, **kwargs) < 0, axis=0) >= 14


def _find_wrapping_mask(original_mask, func, mask_shape, voxel_size, **kwargs) -> ndarray:
    """Find the wrapping mask for an original_mask and a function.
    This is a private function that should not be directly used."""
    # Contributors are discouraged from modifying this function.
    wrapping_mask = zeros_like(original_mask)
    is_3d = original_mask.shape[2] > 1
    end_x, end_y, end_z = np.shape(original_mask)  # The end variable will be used like the end keyword in Matlab.
    end_x, end_y, end_z = end_x - 1, end_y - 1, end_z - 1  # To satisfy zero-based indexing.
    num_step_list = ceil(array((0.05, 0.10, 0.15, 0.20, 0.25, 0.40, 0.50)) * max(mask_shape),
                         casting='unsafe', dtype=int)
    wrapping_failed_msg = ('Could not find a wrapping mask for the %s. '
                           'This is highly improbable and the package author would like to check the model.')

    # Edges or Faces:
    # Note that all =0 and all =end are the same, except for the fancy indexing parts.
    # Check the edge or face where x=0.
    if any(original_mask[0, :, :]):
        wrapping_failed = True
        for num_step in num_step_list:
            x, y, z = np.ogrid[-num_step:0, 0:mask_shape[1], 0:mask_shape[2]]
            new_mask = is_voxel_inside(x, y, z, voxel_size, func, **kwargs)
            if not any(new_mask[0, :, :]):
                wrapping_mask[-num_step:, :, :] = (
                    logical_or(wrapping_mask[-num_step:, :, :], new_mask))
                wrapping_failed = False
                break
        if wrapping_failed:
            raise RuntimeError(wrapping_failed_msg % 'edge or face where x=0')
    # Check edge or face where y=0.
    if any(original_mask[:, 0, :]):
        wrapping_failed = True
        for num_step in num_step_list:
            x, y, z = np.ogrid[0:mask_shape[0], -num_step:0, 0:mask_shape[2]]
            new_mask = is_voxel_inside(x, y, z, voxel_size, func, **kwargs)
            if not any(new_mask[:, 0, :]):
                wrapping_mask[:, -num_step:, :] = (
                    logical_or(wrapping_mask[:, -num_step:, :], new_mask))
                wrapping_failed = False
                break
        if wrapping_failed:
            raise RuntimeError(wrapping_failed_msg % 'edge or face where y=0')
    # Check the face where z=0.
    if is_3d and any(original_mask[:, :, 0]):  # Only for 3D arrays.
        wrapping_failed = True
        for num_step in num_step_list:
            x, y, z = np.ogrid[0:mask_shape[0], 0:mask_shape[1], -num_step:0]
            new_mask = is_voxel_inside(x, y, z, voxel_size, func, **kwargs)
            if not any(new_mask[:, :, 0]):
                wrapping_mask[:, :, -num_step:] = (
                    logical_or(wrapping_mask[:, :, -num_step:], new_mask))
                wrapping_failed = False
                break
        if wrapping_failed:
            raise RuntimeError(wrapping_failed_msg % 'edge or face where z=0')
    # Check edge or face where x=end.
    if any(original_mask[end_x, :, :]):
        wrapping_failed = True
        for num_step in num_step_list:
            x, y, z = np.ogrid[end_x:end_x + num_step, 0:mask_shape[1], 0:mask_shape[2]]
            new_mask = is_voxel_inside(x, y, z, voxel_size, func, **kwargs)
            if not any(new_mask[np.shape(new_mask)[0] - 1, :, :]):
                wrapping_mask[0:num_step, :, :] = (logical_or(wrapping_mask[0:num_step, :, :], new_mask))
                wrapping_failed = False
                break
        if wrapping_failed:
            raise RuntimeError(wrapping_failed_msg % 'edge or face where x=end')
    # Check edge or face where y=end.
    if any(original_mask[:, end_y, :]):
        wrapping_failed = True
        for num_step in num_step_list:
            x, y, z = np.ogrid[0:mask_shape[0], end_y:end_y + num_step, 0:mask_shape[2]]
            new_mask = is_voxel_inside(x, y, z, voxel_size, func, **kwargs)
            if not any(new_mask[:, np.shape(new_mask)[1] - 1, :]):
                wrapping_mask[:, 0:num_step, :] = (logical_or(wrapping_mask[:, 0:num_step, :], new_mask))
                wrapping_failed = False
                break
        if wrapping_failed:
            raise RuntimeError(wrapping_failed_msg % 'edge or face where y=end')
    # Check the face where z=end.
    if is_3d and any(original_mask[:, :, end_z]):  # Only for 3D arrays.
        wrapping_failed = True
        for num_step in num_step_list:
            x, y, z = np.ogrid[0:mask_shape[0], 0:mask_shape[1], end_z:end_z + num_step]
            new_mask = is_voxel_inside(x, y, z, voxel_size, func, **kwargs)
            if not any(new_mask[:, :, np.shape(new_mask)[2] - 1]):
                wrapping_mask[:, :, 0:num_step] = (logical_or(wrapping_mask[:, :, 0:num_step], new_mask))
                wrapping_failed = False
                break
        if wrapping_failed:
            raise RuntimeError(wrapping_failed_msg % 'face where z=end')

    # Vertices:
    # Note that each vertex is a combination of the indexing of two edge/face logics.
    # Check the vertex where x=0 and y=0.
    if any(original_mask[0, :, :]) or any(original_mask[:, 0, :]):
        wrapping_failed = True
        for num_step in num_step_list:
            x, y, z = np.ogrid[-num_step:0, -num_step:0, 0:mask_shape[2]]
            new_mask = is_voxel_inside(x, y, z, voxel_size, func, **kwargs)
            if not (any(new_mask[0, :, :]) or any(new_mask[:, 0, :])):
                wrapping_mask[-num_step:, -num_step:, :] = (
                    logical_or(wrapping_mask[-num_step:, -num_step:, :], new_mask))
                wrapping_failed = False
                break
        if wrapping_failed:
            raise RuntimeError(wrapping_failed_msg % 'vertex where x=0 and y=0')
    # Check the vertex where x=0 and y=end.
    if any(original_mask[0, :, :]) or any(original_mask[:, end_y, :]):
        wrapping_failed = True
        for num_step in num_step_list:
            x, y, z = np.ogrid[-num_step:0, end_y:end_y + num_step, 0:mask_shape[2]]
            new_mask = is_voxel_inside(x, y, z, voxel_size, func, **kwargs)
            if not (any(new_mask[0, :, :]) or any(new_mask[:, np.shape(new_mask)[1] - 1, :])):
                wrapping_mask[-num_step:, 0:num_step, :] = (
                    logical_or(wrapping_mask[-num_step:, 0:num_step, :], new_mask))
                wrapping_failed = False
                break
        if wrapping_failed:
            raise RuntimeError(wrapping_failed_msg % 'vertex where x=0 and y=0')
    # Check the vertex where x=end and y=0.
    if any(original_mask[end_x, :, :]) or any(original_mask[:, 0, :]):
        wrapping_failed = True
        for num_step in num_step_list:
            x, y, z = np.ogrid[end_x:end_x + num_step, -num_step:0, 0:mask_shape[2]]
            new_mask = is_voxel_inside(x, y, z, voxel_size, func, **kwargs)
            if not (any(new_mask[np.shape(new_mask)[0] - 1, :, :]) or any(new_mask[:, 0, :])):
                wrapping_mask[0:num_step, -num_step:, :] = (
                    logical_or(wrapping_mask[0:num_step, -num_step:, :], new_mask))
                wrapping_failed = False
                break
        if wrapping_failed:
            raise RuntimeError(wrapping_failed_msg % 'vertex where x=end and y=0')
    # Check the vertex where x=end and y=end.
    if any(original_mask[end_x, :, :]) or any(original_mask[:, end_y, :]):
        wrapping_failed = True
        for num_step in num_step_list:
            x, y, z = np.ogrid[end_x:end_x + num_step, end_y:end_y + num_step, 0:mask_shape[2]]
            new_mask = is_voxel_inside(x, y, z, voxel_size, func, **kwargs)
            if not (any(new_mask[np.shape(new_mask)[0] - 1, :, :]) or any(new_mask[:, np.shape(new_mask)[1] - 1, :])):
                wrapping_mask[0:num_step, 0:num_step, :] = (
                    logical_or(wrapping_mask[0:num_step, 0:num_step, :], new_mask))
                wrapping_failed = False
                break
        if wrapping_failed:
            raise RuntimeError(wrapping_failed_msg % 'vertex where x=end and y=end')

    # For 2D models, we are done.
    # A return statement is used instead of a conditional to reduce the indent and improve readability.
    if not is_3d:
        return wrapping_mask

    # Check the vertex where x=0 and z=0.
    if any(original_mask[0, :, :]) or any(original_mask[:, :, 0]):
        wrapping_failed = True
        for num_step in num_step_list:
            x, y, z = np.ogrid[-num_step:0, 0:mask_shape[1], -num_step:0]
            new_mask = is_voxel_inside(x, y, z, voxel_size, func, **kwargs)
            if not (any(new_mask[0, :, :]) or any(new_mask[:, :, 0])):
                wrapping_mask[-num_step:, :, -num_step:] = (
                    logical_or(wrapping_mask[-num_step:, :, -num_step:], new_mask))
                wrapping_failed = False
                break
        if wrapping_failed:
            raise RuntimeError(wrapping_failed_msg % 'vertex where x=0 and z=0')
    # Check the vertex where x=0 and z=end.
    if any(original_mask[0, :, :]) or any(original_mask[:, :, end_z]):
        wrapping_failed = True
        for num_step in num_step_list:
            x, y, z = np.ogrid[-num_step:0, 0:mask_shape[1], end_z:end_z + num_step]
            new_mask = is_voxel_inside(x, y, z, voxel_size, func, **kwargs)
            if not (any(new_mask[0, :, :]) or any(new_mask[:, :, np.shape(new_mask)[2] - 1])):
                wrapping_mask[-num_step:, :, 0:num_step] = (
                    logical_or(wrapping_mask[-num_step:, :, 0:num_step], new_mask))
                wrapping_failed = False
                break
        if wrapping_failed:
            raise RuntimeError(wrapping_failed_msg % 'vertex where x=0 and z=end')
    # Check the vertex where x=end and z=0.
    if any(original_mask[end_x, :, :]) or any(original_mask[:, :, 0]):
        wrapping_failed = True
        for num_step in num_step_list:
            x, y, z = np.ogrid[end_x:end_x + num_step, 0:mask_shape[1], -num_step:0]
            new_mask = is_voxel_inside(x, y, z, voxel_size, func, **kwargs)
            if not (any(new_mask[np.shape(new_mask)[0] - 1, :, :]) or any(new_mask[:, :, 0])):
                wrapping_mask[0:num_step, :, -num_step:] = (
                    logical_or(wrapping_mask[0:num_step, :, -num_step:], new_mask))
                wrapping_failed = False
                break
        if wrapping_failed:
            raise RuntimeError(wrapping_failed_msg % 'vertex where x=end and z=0')
    # Check the vertex where x=end and z=end.
    if any(original_mask[end_x, :, :]) or any(original_mask[:, :, end_z]):
        wrapping_failed = True
        for num_step in num_step_list:
            x, y, z = np.ogrid[end_x:end_x + num_step, 0:mask_shape[1], end_z:end_z + num_step]
            new_mask = is_voxel_inside(x, y, z, voxel_size, func, **kwargs)
            if not (any(new_mask[np.shape(new_mask)[0] - 1, :, :]) or any(new_mask[:, :, np.shape(new_mask)[2] - 1])):
                wrapping_mask[0:num_step, :, 0:num_step] = (
                    logical_or(wrapping_mask[0:num_step, :, 0:num_step], new_mask))
                wrapping_failed = False
                break
        if wrapping_failed:
            raise RuntimeError(wrapping_failed_msg % 'vertex where x=end and z=end')
    # Check the vertex where y=0 and z=0.
    if any(original_mask[:, 0, :]) or any(original_mask[:, :, 0]):
        wrapping_failed = True
        for num_step in num_step_list:
            x, y, z = np.ogrid[0:mask_shape[0], -num_step:0, -num_step:0]
            new_mask = is_voxel_inside(x, y, z, voxel_size, func, **kwargs)
            if not (any(new_mask[:, 0, :]) or any(new_mask[:, :, 0])):
                wrapping_mask[:, -num_step:, -num_step:] = (
                    logical_or(wrapping_mask[:, -num_step:, -num_step:], new_mask))
                wrapping_failed = False
                break
        if wrapping_failed:
            raise RuntimeError(wrapping_failed_msg % 'vertex where y=0 and z=0')
    # Check the vertex where y=0 and z=end.
    if any(original_mask[:, 0, :]) or any(original_mask[:, :, end_z]):
        wrapping_failed = True
        for num_step in num_step_list:
            x, y, z = np.ogrid[0:mask_shape[0], -num_step:0, end_z:end_z + num_step]
            new_mask = is_voxel_inside(x, y, z, voxel_size, func, **kwargs)
            if not (any(new_mask[:, 0, :]) or any(new_mask[:, :, np.shape(new_mask)[2] - 1])):
                wrapping_mask[:, -num_step:, 0:num_step] = (
                    logical_or(wrapping_mask[:, -num_step:, 0:num_step], new_mask))
                wrapping_failed = False
                break
        if wrapping_failed:
            raise RuntimeError(wrapping_failed_msg % 'vertex where y=0 and z=end')
    # Check the vertex where y=end and z=0.
    if any(original_mask[:, end_y, :]) or any(original_mask[:, :, 0]):
        wrapping_failed = True
        for num_step in num_step_list:
            x, y, z = np.ogrid[0:mask_shape[0], end_y:end_y + num_step, -num_step:0]
            new_mask = is_voxel_inside(x, y, z, voxel_size, func, **kwargs)
            if not (any(new_mask[:, np.shape(new_mask)[1] - 1, :]) or any(new_mask[:, :, 0])):
                wrapping_mask[:, 0:num_step, -num_step:] = (
                    logical_or(wrapping_mask[:, 0:num_step, -num_step:], new_mask))
                wrapping_failed = False
                break
        if wrapping_failed:
            raise RuntimeError(wrapping_failed_msg % 'vertex where y=end and z=0')
    # Check the vertex where y=end and z=end.
    if any(original_mask[:, end_y, :]) or any(original_mask[:, :, end_z]):
        wrapping_failed = True
        for num_step in num_step_list:
            x, y, z = np.ogrid[0:mask_shape[0], end_y:end_y + num_step, end_z:end_z + num_step]
            new_mask = is_voxel_inside(x, y, z, voxel_size, func, **kwargs)
            if not (any(new_mask[:, np.shape(new_mask)[1] - 1, :]) or any(new_mask[:, :, np.shape(new_mask)[2] - 1])):
                wrapping_mask[:, 0:num_step, 0:num_step] = (
                    logical_or(wrapping_mask[:, 0:num_step, 0:num_step], new_mask))
                wrapping_failed = False
                break
        if wrapping_failed:
            raise RuntimeError(wrapping_failed_msg % 'vertex where y=end and z=end')

    # The following are for the opposite vertex in 3D. Note that is_3d is assumed to be True and is NOT checked.
    # Check the vertex where x=0, y=0, and z=0.
    if any(original_mask[0, :, :]) or any(original_mask[:, 0, :]) or any(original_mask[:, :, 0]):
        wrapping_failed = True
        for num_step in num_step_list:
            x, y, z = np.ogrid[-num_step:0, -num_step:0, -num_step:0]
            new_mask = is_voxel_inside(x, y, z, voxel_size, func, **kwargs)
            if not (any(new_mask[0, :, :]) or any(new_mask[:, 0, :]) or any(new_mask[:, :, 0])):
                wrapping_mask[-num_step:, -num_step:, -num_step:] = (
                    logical_or(wrapping_mask[-num_step:, -num_step:, -num_step:], new_mask))
                wrapping_failed = False
                break
        if wrapping_failed:
            raise RuntimeError(wrapping_failed_msg % 'vertex where x=0, y=0, and z=0')
    # Check the vertex where x=0, y=0, and z=end.
    if any(original_mask[0, :, :]) or any(original_mask[:, 0, :]) or any(original_mask[:, :, end_z]):
        wrapping_failed = True
        for num_step in num_step_list:
            x, y, z = np.ogrid[-num_step:0, -num_step:0, end_z:end_z + num_step]
            new_mask = is_voxel_inside(x, y, z, voxel_size, func, **kwargs)
            if not (any(new_mask[0, :, :]) or any(new_mask[:, 0, :]) or any(new_mask[:, :, np.shape(new_mask)[2] - 1])):
                wrapping_mask[-num_step:, -num_step:, 0:num_step] = (
                    logical_or(wrapping_mask[-num_step:, -num_step:, 0:num_step], new_mask))
                wrapping_failed = False
                break
        if wrapping_failed:
            raise RuntimeError(wrapping_failed_msg % 'vertex where x=0, y=0, and z=end')
    # Check the vertex where x=0, y=end, and z=0.
    if any(original_mask[0, :, :]) or any(original_mask[:, end_y, :]) or any(original_mask[:, :, 0]):
        wrapping_failed = True
        for num_step in num_step_list:
            x, y, z = np.ogrid[-num_step:0, end_y:end_y + num_step, -num_step:0]
            new_mask = is_voxel_inside(x, y, z, voxel_size, func, **kwargs)
            if not (any(new_mask[0, :, :]) or
                    any(new_mask[:, np.shape(new_mask)[1] - 1, :]) or
                    any(new_mask[:, :, 0])):
                wrapping_mask[-num_step:, 0:num_step, -num_step:] = (
                    logical_or(wrapping_mask[-num_step:, 0:num_step:, -num_step:], new_mask))
                wrapping_failed = False
                break
        if wrapping_failed:
            raise RuntimeError(wrapping_failed_msg % 'vertex where x=0, y=end, and z=0')
    # Check the vertex where x=0, y=end, and z=end.
    if any(original_mask[0, :, :]) or any(original_mask[:, end_y, :]) or any(original_mask[:, :, end_z]):
        wrapping_failed = True
        for num_step in num_step_list:
            x, y, z = np.ogrid[-num_step:0, end_y:end_y + num_step, end_z:end_z + num_step]
            new_mask = is_voxel_inside(x, y, z, voxel_size, func, **kwargs)
            if not (any(new_mask[0, :, :]) or
                    any(new_mask[:, np.shape(new_mask)[1] - 1, :]) or
                    any(new_mask[:, :, np.shape(new_mask)[2] - 1])):
                wrapping_mask[-num_step:, 0:num_step, 0:num_step] = (
                    logical_or(wrapping_mask[-num_step:, 0:num_step, 0:num_step], new_mask))
                wrapping_failed = False
                break
        if wrapping_failed:
            raise RuntimeError(wrapping_failed_msg % 'vertex where x=0, y=end, and z=end')
    # Check the vertex where x=end, y=0, and z=0.
    if any(original_mask[end_x, :, :]) or any(original_mask[:, 0, :]) or any(original_mask[:, :, 0]):
        wrapping_failed = True
        for num_step in num_step_list:
            x, y, z = np.ogrid[end_x:end_x + num_step, -num_step:0, -num_step:0]
            new_mask = is_voxel_inside(x, y, z, voxel_size, func, **kwargs)
            if not (any(new_mask[np.shape(new_mask)[0] - 1, :, :]) or
                    any(new_mask[:, 0, :]) or
                    any(new_mask[:, :, 0])):
                wrapping_mask[0:num_step, -num_step:, -num_step:] = (
                    logical_or(wrapping_mask[0:num_step, -num_step:, -num_step:], new_mask))
                wrapping_failed = False
                break
        if wrapping_failed:
            raise RuntimeError(wrapping_failed_msg % 'vertex where x=end, y=0, and z=0')
    # Check the vertex where x=end, y=0, and z=end.
    if any(original_mask[end_x, :, :]) or any(original_mask[:, 0, :]) or any(original_mask[:, :, end_z]):
        wrapping_failed = True
        for num_step in num_step_list:
            x, y, z = np.ogrid[end_x:end_x + num_step, -num_step:0, end_z:end_z + num_step]
            new_mask = is_voxel_inside(x, y, z, voxel_size, func, **kwargs)
            if not (any(new_mask[np.shape(new_mask)[0] - 1, :, :]) or
                    any(new_mask[:, 0, :]) or
                    any(new_mask[:, :, np.shape(new_mask)[2] - 1])):
                wrapping_mask[0:num_step, -num_step:, 0:num_step] = (
                    logical_or(wrapping_mask[0:num_step, -num_step:, 0:num_step], new_mask))
                wrapping_failed = False
                break
        if wrapping_failed:
            raise RuntimeError(wrapping_failed_msg % 'vertex where x=end, y=0, and z=end')
    # Check the vertex where x=end, y=end, and z=0.
    if any(original_mask[end_x, :, :]) or any(original_mask[:, end_y, :]) or any(original_mask[:, :, 0]):
        wrapping_failed = True
        for num_step in num_step_list:
            x, y, z = np.ogrid[end_x:end_x + num_step, end_y:end_y + num_step, -num_step:0]
            new_mask = is_voxel_inside(x, y, z, voxel_size, func, **kwargs)
            if not (any(new_mask[np.shape(new_mask)[0] - 1, :, :]) or
                    any(new_mask[:, np.shape(new_mask)[1] - 1, :]) or
                    any(new_mask[:, :, 0])):
                wrapping_mask[0:num_step, 0:num_step, -num_step:] = (
                    logical_or(wrapping_mask[0:num_step, 0:num_step:, -num_step:], new_mask))
                wrapping_failed = False
                break
        if wrapping_failed:
            raise RuntimeError(wrapping_failed_msg % 'vertex where x=end, y=end, and z=0')
    # Check the vertex where x=end, y=end, and z=end.
    if any(original_mask[end_x, :, :]) or any(original_mask[:, end_y, :]) or any(original_mask[:, :, end_z]):
        wrapping_failed = True
        for num_step in num_step_list:
            x, y, z = np.ogrid[end_x:end_x + num_step, end_y:end_y + num_step, end_z:end_z + num_step]
            new_mask = is_voxel_inside(x, y, z, voxel_size, func, **kwargs)
            if not (any(new_mask[np.shape(new_mask)[0] - 1, :, :]) or
                    any(new_mask[:, np.shape(new_mask)[1] - 1, :]) or
                    any(new_mask[:, :, np.shape(new_mask)[2] - 1])):
                wrapping_mask[0:num_step, 0:num_step, 0:num_step] = (
                    logical_or(wrapping_mask[0:num_step, 0:num_step, 0:num_step], new_mask))
                wrapping_failed = False
                break
        if wrapping_failed:
            raise RuntimeError(wrapping_failed_msg % 'vertex where x=end, y=end, and z=end')

    return wrapping_mask



"""Classes defining geometrical shapes which can be used to create Boolean masks.

The resulting masks can then be used
for manipulating a :class:`~vcams.voxelpart.VoxelPart` instance
using its :meth:`~vcams.voxelpart.VoxelPart.apply_mask` method.
See the :ref:`predefined-shape` section for a complete explanation
of the basic concepts.
"""
from abc import ABC, abstractmethod
from collections.abc import Iterable
from copy import deepcopy
from itertools import count
from typing import Union

import numpy as np
from numpy import logical_or, ndarray, sin, cos, radians, any, logical_and, full, pi, count_nonzero, \
    prod, isscalar, array

from vcams.mask.function import mask_from_function


class BaseShape(ABC):
    """Abstract base class describing a shape.

    All shapes must inherit from this class.
    Subclasses define their dimensionality using the *dim* class attribute
    which can be either be '2D' or '3D',
    and define the level-set function *func* describing the shape in 3D space.
    It must be compatible with :func:`vcams.mask.function.mask_from_function`.
    """

    def __init__(self):
        self.part_shape = None
        self.voxel_size = None
        self.voxel_volume = None
        self.num_workspace_voxels = None
        self._num_true_voxels = None
        self._real_shape_volume = None
        self._shape_volume_fraction = None

    @property
    @abstractmethod
    def dim(self):
        """The dimensionality of the shape. Must be defined by subclasses to be '2D' or '3D'."""
        pass

    @property
    @abstractmethod
    def analytical_volume(self):
        """Volume of the shape calculated using analytical equations.
        For 2D shapes, surface area is returned. This should only be used as an approximation as
        it does not take into account voxelization errors and contacts with the boundary."""
        pass

    @property
    def num_true_voxels(self):
        """Number of *True* voxels in the voxelized shape (:math:`N_{Shape}^{True}`),
        calculated and set by the :meth:`calculate_mask` function.
        It is initially set to *None* which raises an error when accessed."""
        if self._num_true_voxels is None:
            raise RuntimeError('The number of voxels in the object has not been calculated yet. '
                               'Has calculate_mask() been called yet?')
        else:
            return self._num_true_voxels

    @property
    def real_shape_volume(self):
        """Real volume of the voxelized shape calculated as :math:`N_{Shape}^{True} \\times V_{Voxel}`
        where :math:`N_{Shape}^{True}` is :attr:`num_true_voxels`
        and :math:`V_{Voxel}` is the volume of each voxel.

        The :meth:`calculate_mask` function needs to be called first, otherwise an error is raised."""
        return self.num_true_voxels * self.voxel_volume

    @property
    def real_workspace_volume(self):
        """Real volume of the voxelized workspace (the :class:`~.voxelpart.VoxelPart` object)
        containing the shape.
        It is calculated as :math:`\\prod_{i=x,y,z} N_i L_i`
        where for each dimension :math:`i`, :math:`N_i` is the number of voxels
        and :math:`L_i` is the length of voxels in that direction.

        The :meth:`calculate_mask` function needs to be called first, otherwise an error is raised."""
        if self.voxel_volume is None:
            raise RuntimeError("The object's calculate_mask() has not been called yet.")
        else:
            return self.num_workspace_voxels * self.voxel_volume

    @property
    def shape_volume_fraction(self):
        """Volume fraction of the voxelized shape.
        It is calculated as :math:`\\frac{N_{Shape}^{True}}{N_{Workspace}}`
        where :math:`N_{Shape}^{True}` is the number of *True* voxels
        in the shape (See :attr:`num_true_voxels`)
        and :math:`N_{Workspace}` is the total number of voxels in the workspace.
        Size of each voxel is not relevant to the calculation.

        The :meth:`calculate_mask` function needs to be called first, otherwise an error is raised."""
        if self.num_workspace_voxels is None:
            raise RuntimeError("The object's calculate_mask() has not been called yet.")
        else:
            return self.num_true_voxels / self.num_workspace_voxels

    @abstractmethod
    def func(self, x: float | ndarray, y: float | ndarray, z: float | ndarray,
             boundary_on: bool = False) -> float | ndarray:
        """The level-set function describing the shape in 3D space

        It must be compatible with :func:`vcams.mask.function.mask_from_function`.

        Args:
            x: A float or numpy 1D array of x-coordinates.
            y: A float or numpy 1D array of y-coordinates.
            z: A float or numpy 1D array of z-coordinates.
            boundary_on: A boolean specifying whether shape boundary
            must be considered when evaluating the function.
            The boolean itself is added to the equations as a multiplier and
            evaluates to 0 or 1 which allows for a single expression.

        Returns:
            An array of floats which may be negative, zero, or positive.
            If scalar values are passed, a float is returned instead of an array.
            See :ref:`level-set-functions` for interpretation of the results.
        """

    pass

    def calculate_mask(self, part=None,
                       part_shape: tuple[int, int, int] = None,
                       voxel_size: tuple[float, float, float] = None,
                       wrap_mask: bool = False, boundary_on: bool = False) -> ndarray:
        """Calculate the boolean mask based on this shape.
        This is a wrapper for :func:`vcams.mask.function.mask_from_function`.

        Args:
            part (VoxelPart | None): The *VoxelPart* instance based on which a mask is created.
                                     If *None*, *part_shape* and *voxel_size* must be specified
                                     otherwise they are ignored. Defaults to *None*.
            part_shape: A tuple containing three integers which determine
                        the shape of the returned boolean mask.
                        Defaults to *None* and ignored if *part* is passed.
            voxel_size: A tuple containing three floats which determine the size of a voxel
                        in the x, y, and z directions.
                        Defaults to *None* and ignored if *part* is passed.
            wrap_mask: If set to True, the shape's function is wrapped around
                       the boundaries of the working space.
                       This is useful for periodic structures, but is computationally expensive.
            boundary_on: A boolean specifying whether shape boundary
                         must be considered when evaluating the function.

        Returns:
            A numpy ndarray with a dtype of bool representing the current shape.
        """

        # Set part_shape and voxel_size.
        if part:
            part_shape = part.size
            voxel_size = part.voxel_size
        self.part_shape = part_shape
        self.voxel_size = voxel_size

        # Shape mask is calculated here but is not stored to save memory.
        shape_mask = mask_from_function(part=None, wrap_mask=wrap_mask, mask_shape=part_shape,
                                        voxel_size=voxel_size, func=self.func, boundary_on=boundary_on,
                                        do_log=False)
        self._num_true_voxels = count_nonzero(shape_mask)
        # Calculate shape properties related to voxel_size and part shape.
        if self.dim == '2D':
            self.voxel_volume = prod(voxel_size[:2])
            self.num_workspace_voxels = prod(part_shape[:2])
        else:
            self.voxel_volume = prod(voxel_size)
            self.num_workspace_voxels = prod(part_shape)
        return shape_mask


class ShapeArray:
    """Class for an array of shapes where each shape's parameters and location is predefined.
    The array may contain any number of shapes of any class as long as they
    are subclasses of :class:`.shape.BaseShape` and have the same *dim* attribute.
    """

    def __init__(self, dim: str, part=None,
                 part_shape: tuple[int, int] | tuple[int, int, int] = None,
                 voxel_size: tuple[float, float] | tuple[float, float, float] = None,
                 is_mask_calculation_lazy: bool = True, wrap_mask: bool = False):
        """Constructor for the *ShapeArray* class.

        Args:
            dim: Dimensionality of the shape array which determines the shapes that
                 can be added to the shape array. Valid values are '2D' and '3D'.
            part (VoxelPart | None): The *VoxelPart* instance based on which the shape array is created.
                                     If *None*, *part_shape* and *voxel_size* must be specified,
                                     otherwise they are ignored. Defaults to *None*.
            part_shape: A tuple containing two or three integers which determine
                        the shape of the returned boolean mask.
                        Defaults to *None* and ignored if *part* is passed.
            voxel_size: A tuple containing two or three floats which determine the size of a voxel
                        in the x, y, and z directions.
                        Defaults to *None* and ignored if *part* is passed.
            is_mask_calculation_lazy: If True, the instance's private *_mask* property is updated
                                      only when necessary which greatly improves performance.
                                      Otherwise, it is updated everytime a shape is added to the array.
            wrap_mask: If set to True, the shapes' function is wrapped around
                       the boundaries of the working space, enabling periodic structures.
        """
        if dim.upper() not in ['2D', '3D']:
            raise ValueError("dim can only be one of '2D' or '3D'.")
        self.dim = dim.upper()
        """Dimensionality of the *ShapeArray* instance which determines the shapes that
           can be added to the shape array.
           It is when creating the instance and must not be changed."""

        if part:
            self.part_name = part.name
            self.part_shape = part.size
            self.voxel_size = part.voxel_size
            self.base_mask = (part.data == 0)
            self._part_log_file_path = part._log_file_path  # noqa: PyProtectedMember
        else:
            if (part_shape is None) or (voxel_size is None):
                raise ValueError('The part argument is not specified,'
                                 'therefore both part_shape and voxel_size must be specified.')
            self.part_name = None
            """Name of the part for which the *ShapeArray* instance is created.
            If a pert is not passed, it is set to *None* and it is not used."""
            self.part_shape = array(part_shape)
            """See :meth:`.__init__`."""
            self.voxel_size = array(voxel_size)
            """See :meth:`.__init__`."""
            self.base_mask = full(part_shape, False, dtype=bool)
            """A Boolean mask for the part which contains the background where the shapes are dispersed.
            If a VoxelPart instance is passed, its nonzero elements are considered occupied,
            otherwise a blank mask is used.
            Also in the child :class:`~.shape_dispersion.ShapeDispersionArray` class,
            the boundary pixels, as determined by the *num_bound_pixels* parameter,
            are considered occupied."""
            self._part_log_file_path = None

        # Validate and fix part_shape and voxel_size.
        if self.part_shape.size not in [2, 3]:
            raise ValueError(f'part_shape must have 2 or 3 elements, '
                             f'but it has {self.part_shape.size}.')
        if self.voxel_size.size not in [2, 3]:
            raise ValueError(f'voxel_size must have 2 or 3 elements, '
                             f'but it has {self.voxel_size.size}.')
        if dim.upper() == '2D':
            self.part_shape = self.part_shape[:2]
            self.voxel_size = self.voxel_size[:2]
        else:
            if self.part_shape.size != 3:
                raise ValueError(f'The shape array is defined as 3D, '
                                 f'but part_shape has {self.part_shape.size} elements.')
            if self.voxel_size.size != 3:
                raise ValueError(f'The shape array is defined as 3D, '
                                 f'but voxel_size has {self.voxel_size.size} elements.')
        self.part_volume = prod(self.part_shape * self.voxel_size)

        self.is_mask_calculation_lazy = is_mask_calculation_lazy
        """See :meth:`.__init__`."""
        self.wrap_mask = wrap_mask
        """See :meth:`.__init__`."""
        self._deferred_masks = []
        # Private attribute for masks that are have not been calculated
        # in the _calculate_mask method due to is_mask_calculation_lazy being True.
        # They will be included and this list will be emptied when necessary.
        self._mask = None  # Private attribute for the mask property.
        self._full_mask = None  # Private attribute for the _full_mask property.
        self.shapes = dict()
        """A dictionary containing the shapes in the instance.
        Keys are integer shape IDs, and values are subclasses of :class:`.shape.BaseShape`."""

        self._backup_dict = dict()
        """A dictionary containing the state of the instance
        which is used by :meth:`_backup_state` and :meth:`_restore_state`.
        It should not be used directly."""

    def __len__(self):
        return len(self.shapes)

    id_iter: iter = count()
    """An iterable keeping track of the number of shapes in the ShapeArray."""

    @property
    def mask(self):
        """A Boolean mask representing the union (logical OR) of the shapes in ShapeArray.
        This is guaranteed to be up-to-date.
        """
        if self._deferred_masks or (self._mask is None):
            self._calculate_mask(shape_id=self._deferred_masks)
            # Updates both self._mask and self._full_mask.
        return self._mask

    @property
    def full_mask(self):
        """A Boolean containing both the background and current dispersed shapes.
        It is a union (logical OR) of the ShapeArray's :attr:`base_mask` and :attr:`mask` attributes
        and is stored to reduce repetitive computations. This is guaranteed to be up-to-date.
        """
        if self._deferred_masks or (self._mask is None):
            self._calculate_mask(shape_id=self._deferred_masks)
            # Updates both self._mask and self._full_mask.
        return self._full_mask

    @property
    def shape_array_volume_fraction(self):
        """Ratio of the occupied (*True*) voxels in the instance
        to the instance's total number of voxels.
        This only includes the voxels in :attr:`mask` and the voxels
        in the background (:attr:`base_mask`) are not included."""
        return np.count_nonzero(self.mask) / prod(self.part_shape)

    def _check_shape_class(self, shape):
        """Validate the given shape class or instance.
         Currently, only its dimensionality (*dim* attribute)
         is checked against that of the shape array."""
        if shape.dim != self.dim:
            raise ValueError(
                'The specified shape is %s, but the shape array has been defined for %s shapes.'
                % (shape.dim, self.dim))

    def _calculate_mask(self, shape_id: int | Iterable | None = None):
        """Calculate the Boolean mask for the entire ShapeArray or only the given shapes.

        Args:
            shape_id: An integer ID or iterable of IDs of shapes for which Boolean masks
                      need to be calculated and added to the shape.
                      If set to *None*, all shapes will have their Boolean masks recalculated.
                      Defaults to *None*.
        """

        # Make sure there are shapes in the ShapeArray.
        if len(self) == 0:
            self._mask = None
            self._full_mask = None

        if (self._mask is None) or (shape_id is None):  # All masks need to be calculated.
            recalculate_all = True
            # Create an empty boolean array for the mask.
            self._mask = full(self.part_shape, False, dtype=bool)
            # Get id_list from the shapes dictionary.
            id_list = list(self.shapes.keys())
        else:  # Some masks need to be calculated.
            recalculate_all = False
            if isscalar(shape_id):
                shape_id = (shape_id,)
            id_list = shape_id

        # Either way, we iterate over id_list and call logical_or.
        for i in id_list:
            self._mask = logical_or(self._mask,
                                    self.shapes[i].calculate_mask(
                                        part_shape=self.part_shape, voxel_size=self.voxel_size,
                                        wrap_mask=self.wrap_mask))
            try:
                self._deferred_masks.remove(i)
            except ValueError:
                pass

        # If all is calculated, empty _deferred_masks.
        if recalculate_all:
            self._deferred_masks = []

        # Update self._full_mask.
        self._full_mask = logical_or(self._mask, self.base_mask)

    def _add_shape_obj(self, shape_obj: BaseShape):
        """Add an existing shape object to the ShapeArray."""
        idd = next(self.id_iter)
        shape_obj.id = idd
        self.shapes[idd] = shape_obj
        if not self.is_mask_calculation_lazy:
            self._calculate_mask(shape_id=idd)
        else:
            self._deferred_masks.append(idd)

    def _backup_state(self):
        """Backup the state of the instance, so it can be restored later.

        This is a private function intended for use by the methods of
        the :class:`~.shape_dispersion.ShapeDispersionArray` subclass
        and should not be directly called by the users.
        In case of a mistake in the *ShapeArray*, start from scratch."""
        self._backup_dict = dict()  # Remove previous backup.
        self._backup_dict['id_iter'] = deepcopy(self.id_iter)
        self._backup_dict['wrap_mask'] = self.wrap_mask
        self._backup_dict['base_mask'] = ndarray.copy(self.base_mask)
        self._backup_dict['shapes'] = deepcopy(self.shapes)
        if len(self.shapes) == 0:  # Set mask and full_mask the same as self.__init__().
            self._backup_dict['mask'] = None
            self._backup_dict['full_mask'] = None
        else:
            self._backup_dict['mask'] = ndarray.copy(self.mask)
            self._backup_dict['full_mask'] = ndarray.copy(self.full_mask)
        if len(self._deferred_masks) != 0:  # It is emptied by mask() and full_mask().
            raise RuntimeError("The instance's _deferred_masks attribute"
                               "is not an empty list, but it should be.")

    def _restore_state(self, backup_dict=None):
        """Restore the state of the instance to the previous backup.

        This is a private function intended for use by the methods of
        the :class:`~.shape_dispersion.ShapeDispersionArray` subclass
        and should not be directly called by the users.
        In case of a mistake in the *ShapeArray*, start from scratch.

        Args:
            backup_dict: The :attr:`_backup_dict` dictionary.
                         If set to *None*, the instance's *_backup_dict* is used.
                         Otherwise, another dictionary can be passed which should
                         ideally be a deep copy of an existing _backup_dict.
        """
        if backup_dict is None:
            backup_dict = self._backup_dict
        if not backup_dict:
            raise RuntimeError("The instance's _backup_dict property is empty."
                               "Has _backup_state() been called before? This may also happen after "
                               "a successful run of ShapeDispersionArray.disperse_shapes().")
        self.id_iter = deepcopy(backup_dict['id_iter'])
        self.wrap_mask = backup_dict['wrap_mask']
        self.base_mask = ndarray.copy(backup_dict['base_mask'])
        if backup_dict['mask'] is None:
            self._mask = None
        else:
            self._mask = ndarray.copy(backup_dict['mask'])
        if backup_dict['mask'] is None:
            self._full_mask = None
        else:
            self._full_mask = ndarray.copy(backup_dict['full_mask'])
        self.shapes = deepcopy(backup_dict['shapes'])

    def add_shape(self, cls, intersect_ok: bool = True, **kwargs) -> bool:
        """Add a shape to the ShapeArray using its class,
        while checking for intersection with other shapes.
        The arguments are passed as *\\*\\*kwargs* and the shape ID is set automatically.

        Args:
            cls: The shape class that should be added.
                 It should be a subclass of :class:`.shape.BaseShape`.
                 Arguments are passed as *kwargs*.
            intersect_ok: Whether intersection of the new shape with the rest is OK.
                          If *True*, shapes are added without checking for intersection.
                          If *False*, the function checks the union (logical OR)
                          of the new shape's Boolean mask and the ShapeArray's :attr:`full_mask`
                          and if they intersect, the shape is not added.
                          Defaults to *True*.
            **kwargs: A dictionary where the keys are the arguments for the shape class
                      and the values are either scalar or dispersion objects
                      from the :mod:`.shape_dispersion` module.

        Returns:
            *True* if the shape was added and *False* otherwise.
        """

        # Validate the shape class.
        self._check_shape_class(cls)
        # Create the new shape object.
        new_shape_obj = cls(id=-1, **kwargs)

        if intersect_ok:
            # Just add the shape.
            self._add_shape_obj(shape_obj=new_shape_obj)
            return True
        else:
            # Calculate the new shape's mask and
            # add only if the shape does not intersect existing shapes or the background.
            new_shape_mask = new_shape_obj.calculate_mask(
                part_shape=self.part_shape, voxel_size=self.voxel_size,
                wrap_mask=self.wrap_mask, boundary_on=True)
            if any(logical_and(self.full_mask, new_shape_mask)):
                # If they intersect, return False for an unsuccessful operation.
                # The shape will be discarded.
                return False
            else:
                # If they don't intersect, return True for a successful operation,
                # and add the shape to the array.
                self._add_shape_obj(shape_obj=new_shape_obj)
                return True

    def remove_shape(self, id_list: int | Iterable[int]):
        """Remove shapes from the :class:`ShapeDispersionArray` instance and update the masks.

        Args:
            id_list: Scalar ID of the shape to be removed, or an iterable of the IDs.
        """

        # Validate id_list.
        if isscalar(id_list):
            id_list = (id_list,)
        for shape_id in id_list:
            if shape_id not in self.shapes.keys():
                raise ValueError(f'{shape_id} is not a valid shape ID.')

        # Delete all shapes from id_list from the shapes dictionary.
        for shape_id in id_list:
            del self.shapes[shape_id]

        # Recalculate self._mask and self._full_mask.
        self._calculate_mask()  # shape_id defaults to None which recalculates all.


class Circle(BaseShape):
    """Class describing a 2D Circle with the implicit equation:

    .. math::
       :label: shape-circle-eq

       (x-x_c)^2 + (y-y_c)^2 - r^2 = 0
    """

    def __init__(self, id: int, xc: float, yc: float, r: float, br: float = 0):
        """
        Args:
            id: ID of the shape which should be must be unique.
            xc: x-coordinate of the center of the circle. It must be positive.
            yc: y-coordinate of the center of the circle. It must be positive.
            r: Radius of the circle. It must be positive.
            br: Radial boundary added to *r* when dispersing the shape. Defaults to 0.
        """
        super().__init__()
        self.id = id
        self.xc = xc
        self.yc = yc
        self.br = br
        if r <= 0:
            raise ValueError(f'r must be positive but is {r:.6f}')
        else:
            self.r = r

    dim: str = '2D'
    """This class attribute means that shape can be used for 2D models."""

    @property
    def analytical_volume(self):
        """Volume of the circle calculated using the analytical equation :math:`\\pi r^2`.
        Because of the voxel meshing, actual volume will be different.
        This should only be used as an approximation as
        it does not take into account voxelization errors and contacts with the boundary.
        Note that the word volume is used instead of surface to simplify the interface."""
        return pi * self.r ** 2

    def func(self, x: float | ndarray, y: float | ndarray, z: float | ndarray,
             boundary_on: bool = False) -> float | ndarray:
        return (x - self.xc) ** 2 + (y - self.yc) ** 2 - (self.r + boundary_on * self.br) ** 2


class Sphere(BaseShape):
    """Class describing a 3D Sphere with the implicit equation:

    .. math::
       :label: shape-sphere-eq

       (x-x_c)^2 + (y-y_c)^2 + (z-z_c)^2 - r^2 = 0
    """

    def __init__(self, id, xc: float, yc: float, zc: float, r: float, br: float = 0):
        """
        Args:
            id: ID of the shape which should be must be unique.
            xc: x-coordinate of the center of the sphere. It must be positive.
            yc: y-coordinate of the center of the sphere. It must be positive.
            zc: y-coordinate of the center of the sphere. It must be positive.
            r: Radius of the sphere. It must be positive.
            br: Radial boundary added to *r* when dispersing the shape. Defaults to 0.
        """
        super().__init__()
        self.id = id
        self.xc = xc
        self.yc = yc
        self.zc = zc
        self.br = br
        if r <= 0:
            raise ValueError(f'r must be positive but is {r:.6f}')
        else:
            self.r = r

    dim: str = '3D'
    """This class attribute means that shape can be used for 3D models."""

    @property
    def analytical_volume(self):
        """Volume of the sphere calculated using the analytical equation
        :math:`\\frac{4}{3} \\pi r ^3`.
        This should only be used as an approximation as
        it does not take into account voxelization errors and contacts with the boundary."""
        return (4 * pi * self.r ** 3) / 3

    def func(self, x: float | ndarray, y: float | ndarray, z: float | ndarray,
             boundary_on: bool = False) -> float | ndarray:
        return (x - self.xc) ** 2 + (y - self.yc) ** 2 + (z - self.zc) ** 2 \
               - (self.r + (boundary_on * self.br)) ** 2


class Cylinder(BaseShape):
    """Class describing a 3D Cylinder with the axis in one of x, y, or z directions.
    """

    def __init__(self, id, dir: str, a: float, b: Union[float, None], c: Union[float, None], r: Union[float, None]):
        """
        Args:
            id:  ID of the shape which should be must be unique.
            dir: Direction of the axis. Can be 'x', 'y', or 'z'.
            a:   x-coordinate of the center of the cylinder or *None* if it's in the direction of axis.
            b:   y-coordinate of the center of the cylinder or *None* if it's in the direction of axis.
            c:   z-coordinate of the center of the cylinder or *None* if it's in the direction of axis.
            r:   Radius of the cylinder.
        """
        super().__init__()
        if dir.lower() not in ('x', 'y', 'z'):
            raise ValueError("dir must be one of 'x', 'y', or 'z'.")
        self.dir = dir.lower()
        self.id = id
        self.a = a
        self.b = b
        self.c = c
        self.r = r

    dim: str = '3D'
    """This class attribute means that shape can be used for 3D models."""

    @property
    def analytical_volume(self):
        raise NotImplementedError('This functionality has not been implemented.')

    def func(self, x: float | ndarray, y: float | ndarray, z: float | ndarray,
             boundary_on: bool = False) -> float | ndarray:
        if self.dir == 'x':
            return (x - x) + (y - self.b) ** 2 + (z - self.c) ** 2 - self.r ** 2
        elif self.dir == 'y':
            return (x - self.a) ** 2 + (y - y) + (z - self.c) ** 2 - self.r ** 2
        elif self.dir == 'z':
            return (x - self.a) ** 2 + (y - self.b) ** 2 + (z - z) - self.r ** 2
        else:
            raise RuntimeError("self.dir is equal to '%s' which is not valid and"
                               "should have been caught in the constructor."
                               "Please contact the author." % self.dir)


class Ellipse(BaseShape):
    """Class describing a 2D Ellipse with the implicit equation:

    .. math::
       :label: shape-ellipse_eq

       \\frac{((x-x_c)\\cos(\\alpha) - (y-y_c)\\sin(\\alpha))^2}{a^2}
       + \\frac{((x-x_c) \\sin(\\alpha) + (y-y_c) \\cos(\\alpha))^2}{b^2}
       - 1 = 0

    Where :math:`(x_c, y_c)` is the center of the ellipse,
    :math:`a` and :math:`b` are the length of the semi-axes along the unrotated x and y axes,
    and :math:`\\alpha` is the rotation of the ellipse around the z-axis.

    Note that :math:`\\alpha` is counterclockwise when viewed in the direction of the z-axis,
    but this is not the default view in most viewers. This means that it may be viewed as clockwise.

    This formula is a simple form of the equations developed for :class:`~Ellipsoid`.
    See the docs for that class for a general ellipsoid.
    """

    def __init__(self, id: int, alpha: float, xc: float, yc: float,
                 a: float, b: float, ba: float = 0, bb: float = 0):
        """
        Args:
            id: ID of the shape which should be must be unique.
            alpha: Counterclockwise rotation of the ellipse around the z-axis.
                   It must be in the range [0, 360] in degrees.
            xc: x-coordinate of the center of the ellipse where semi-axes meet.
            yc: y-coordinate of the center of the ellipse where semi-axes meet.
            a: Semi-axis of the ellipse along the unrotated x-axis. It must be positive.
            b: Semi-axis of the ellipse along the unrotated y-axis. It must be positive.
            ba: Boundary added to *a* when dispersing the shape. Defaults to 0.
            bb: Boundary added to *b* when dispersing the shape. Defaults to 0.
        """
        super().__init__()
        self.id = id
        self.xc = xc
        self.yc = yc
        self.ba = ba
        self.bb = bb
        if a <= 0:
            raise ValueError(f'a must be positive but is {a:.6f}')
        else:
            self.a = a
        if b <= 0:
            raise ValueError(f'b must be positive but is {b:.6f}')
        else:
            self.b = b
        # print(alpha)
        if alpha > 360 or alpha < 0:
            raise ValueError(f'alpha must be in the range [0, 360], but is {alpha:.6f}')
        else:
            self.alpha = radians(alpha)

    dim: str = '2D'
    """This class attribute means that shape can be used for 2D models."""

    @property
    def analytical_volume(self):
        """Volume of the ellipse calculated using the analytical equation :math:`\\pi a b`,
        where :math:`a` and :math:`b` are the semi-axes.
        This should only be used as an approximation as
        it does not take into account voxelization errors and contacts with the boundary.
        Note that the word volume is used instead of surface to simplify the interface."""
        return pi * self.a * self.b

    def func(self, x: float | ndarray, y: float | ndarray, z: float | ndarray,
             boundary_on: bool = False) -> float | ndarray:
        return (
                ((((x - self.xc) * cos(self.alpha) - (y - self.yc) * sin(self.alpha)) ** 2)
                 / (self.a + boundary_on * self.ba) ** 2)
                + ((((x - self.xc) * sin(self.alpha) + (y - self.yc) * cos(self.alpha)) ** 2)
                   / (self.b + boundary_on * self.bb) ** 2)
                - 1)


class EllipseFromAspectRatio(Ellipse):
    """Class describing a 2D Ellipse defined using one axis and the ellipse's aspect ratio.

    See :class:`.Ellipse` for the parent class.
    """

    def __init__(self, id: int, alpha: float, xc: float, yc: float,
                 a: float, aspect_ratio: float, ba: float = 0, bb: float = 0):
        """
        Args:
            id: ID of the shape which should be must be unique.
            alpha: Counterclockwise rotation of the ellipse around the z-axis.
                   It must be in the range [0, 360] in degrees.
            xc: x-coordinate of the center of the ellipse where semi-axes meet.
            yc: y-coordinate of the center of the ellipse where semi-axes meet.
            a: Semi-axis of the ellipse along the unrotated x-axis. It must be positive.
            aspect_ratio: The ratio b/a.
                          The attribute b (the semi-axes along the unrotated y-axis) is
                          calculated as aspect_ratio × a.
            ba: Boundary added to *a* when dispersing the shape. Defaults to 0.
            bb: Boundary added to *b* when dispersing the shape. Defaults to 0.
        """

        # Everything is checked in Ellipse.__init__().
        # We only need to check a and calculate b and then pass it to that function.
        if a <= 0:
            raise ValueError(f'a must be positive but is {a:.6f}')
        if aspect_ratio < 0:
            raise ValueError(f'aspect_ratio must be positive but is {aspect_ratio:.6f}')
        else:
            self.aspect_ratio = aspect_ratio
            b = a * aspect_ratio
        super().__init__(id=id, alpha=alpha, xc=xc, yc=yc,
                         a=a, b=b, ba=ba, bb=bb)


class Ellipsoid(BaseShape):
    """Class describing a triaxial Ellipsoid with rotation and translation.

    The implicit equation for an unrotated ellipsoid in the center is:

    .. math::
       :label: shape-ellipsoid-simple

       \\frac{x}{a^2} + \\frac{y}{b^2} + \\frac{z}{c^2} - 1 = 0

    To rotate the ellipsoid, we need to transform the :math:`xyz` coordinates to the new :math:`x'y'z'` system.
    Three rotations must be applied in the following order:

      - A rotation :math:`\\boldsymbol{R_x}(\\gamma)` about the ellipsoid's x-axis.
      - A rotation :math:`\\boldsymbol{R_y}(\\beta)` about the ellipsoid's y-axis.
      - A rotation :math:`\\boldsymbol{R_z}(\\alpha)` about the ellipsoid's z-axis.

    The combination of these rotations is an intrinsic rotation
    whose Tait–Bryan angles are :math:`\\gamma`, :math:`\\beta`, and :math:`\\alpha`.
    We can represent the complete rotation as:

    .. math::
       :label: shape-ellipsoid-rotation

       \\begin{aligned}
       \\boldsymbol{R} &= \\boldsymbol{R_z}(\\alpha)\\boldsymbol{R_y}(\\beta)\\boldsymbol{R_x}(\\gamma)\\\\[12pt]
       &=
       \\begin{bmatrix}
         \\cos(\\alpha) & -\\sin(\\alpha)  & 0 \\\\
         \\sin(\\alpha) &  \\cos(\\alpha)  & 0 \\\\
         0              & 0                & 1
       \\end{bmatrix}
       \\begin{bmatrix}
         \\cos(\\beta)  & 0 & \\sin(\\beta) \\\\
         0              & 1 & 0             \\\\
         -\\sin(\\beta) & 0 & \\cos(\\beta)
       \\end{bmatrix}
       \\begin{bmatrix}
         1 & 0              & 0               \\\\
         0 & \\cos(\\gamma) & -\\sin(\\gamma) \\\\
         0 & \\sin(\\gamma) &  \\cos(\\gamma)
       \\end{bmatrix} \\\\[12pt]
       &=
       \\begin{bmatrix}
         \\cos(\\alpha)\\cos(\\beta) & \\cos(\\alpha)\\sin(\\beta)\\sin(\\gamma)-\\sin(\\alpha)\\cos(\\gamma) & \\cos(\\alpha)\\sin(\\beta)\\cos(\\gamma)+\\sin(\\alpha)\\sin(\\gamma) \\\\
         \\sin(\\alpha)\\cos(\\beta) & \\sin(\\alpha)\\sin(\\beta)\\sin(\\gamma)+\\cos(\\alpha)\\cos(\\gamma) & \\sin(\\alpha)\\sin(\\beta)\\cos(\\gamma)-\\cos(\\alpha)\\sin(\\gamma) \\\\
         -\\sin(\\beta)      & \\cos(\\beta)\\sin(\\gamma)                    & \\cos(\\beta)\\cos(\\gamma)
       \\end{bmatrix}\\\\
       \\end{aligned}

    The main goal is to find the state of a point :math:`P(x,y,z)` with regards to the transformed ellipsoid.
    To do that, the coordinates of :math:`P` must undergo the same transformation as the ellipsoid.
    This means that first it must be translated by the vector :math:`(x_c, y_c, z_c)`,
    and then the rotation :math:`\\boldsymbol{R}` must be applied to it.
    The formula for this transformation is:

    .. math::
       :label: shape-ellipsoid-pdot-transform

       \\begin{bmatrix}x'\\\\y'\\\\z'\\end{bmatrix}
       =\\boldsymbol{R}(\\alpha, \\beta, \\gamma)\\begin{bmatrix}x-x_c\\\\y-y_c\\\\z-z_c\\end{bmatrix}

    Now that we have :math:`P'(x',y',z')`, we can rewrite Eq. :eq:`shape-ellipsoid-simple` for :math:`P'`:

    .. math::
       :label: shape-ellipsoid-pdot-eq

       \\frac{x'}{a^2} + \\frac{y'}{b^2} + \\frac{z'}{c^2} - 1 = 0

    The actual implementation is a little different.
    Using Eqs. :eq:`shape-ellipsoid-rotation` and :eq:`shape-ellipsoid-pdot-transform`,
    and MATLAB's Symbolic Math Toolbox,
    the expressions for each of :math:`x'`, :math:`y'`, and :math:`z'` are found.
    This allows us to use the transformed coordinates :math:`P'(x',y',z')`
    to evaluate Eq. :eq:`shape-ellipsoid-pdot-eq`.

    The reason for this different approach lies in the complex vectorization
    performed in :mod:`~vcams.mask.function` module.
    """

    def __init__(self, id: int, alpha: float, beta: float, gamma: float,
                 xc: float, yc: float, zc: float,
                 a: float, b: float, c: float,
                 ba: float = 0, bb: float = 0, bc: float = 0):
        """
        Args:
            id: ID of the shape which should be must be unique.
            alpha: Rotation of the ellipsoid about its z-axis. It must be in the range [0, 360] degrees.
            beta: Rotation of the ellipsoid around its y-axis. It must be in the range [0, 180] degrees.
            gamma: Rotation of the ellipsoid around its x-axis. It must be in the range [0, 360] degrees.
            xc: x-coordinate of the center of the ellipsoid where semi-axes meet.
            yc: y-coordinate of the center of the ellipsoid where semi-axes meet.
            zc: z-coordinate of the center of the ellipsoid where semi-axes meet.
            a: Semi-axis of the ellipsoid along the unrotated x-axis. It must be positive.
            b: Semi-axis of the ellipsoid along the unrotated y-axis. It must be positive.
            c: Semi-axis of the ellipsoid along the unrotated z-axis. It must be positive.
            ba: Boundary added to *a* when dispersing the shape. Defaults to 0.
            bb: Boundary added to *b* when dispersing the shape. Defaults to 0.
            bc: Boundary added to *c* when dispersing the shape. Defaults to 0.
        """
        super().__init__()
        self.id = id
        self.xc = xc
        self.yc = yc
        self.zc = zc
        self.ba = ba
        self.bb = bb
        self.bc = bc
        if a <= 0:
            raise ValueError(f'a must be positive but is {a:.6f}')
        else:
            self.a = a
        if b <= 0:
            raise ValueError(f'b must be positive but is {b:.6f}')
        else:
            self.b = b
        if c <= 0:
            raise ValueError(f'c must be positive but is {c:.6f}')
        else:
            self.c = c
        if alpha > 360 or alpha < 0:
            raise ValueError(f'alpha must be in the range [0, 360], but is {alpha:.6f}')
        else:
            self.alpha = radians(alpha)
        if beta > 180 or beta < 0:
            raise ValueError(f'beta must be in the range [0, 180], but is {beta:.6f}')
        else:
            self.beta = radians(beta)
        if gamma > 360 or gamma < 0:
            raise ValueError(f'gamma must be in the range [0, 360], but is {gamma:.6f}')
        else:
            self.gamma = radians(gamma)

    dim: str = '3D'
    """This class attribute means that shape can be used for 3D models."""

    @property
    def analytical_volume(self):
        """Volume of the ellipsoid calculated using the analytical equation
        :math:`\\frac{4}{3}\\pi a b c`,
        where :math:`a`, :math:`b`, and :math:`c` are the semi-axes.
        This should only be used as an approximation as
        it does not take into account voxelization errors and contacts with the boundary."""
        return (4 * pi * self.a * self.b * self.c) / 3

    def func(self, x: float | ndarray, y: float | ndarray, z: float | ndarray,
             boundary_on: bool = False) -> float | ndarray:
        # Apply the translation.
        xx = x - self.xc
        yy = y - self.yc
        zz = z - self.zc

        # Apply the rotation.
        xxx = (xx * cos(self.alpha) * cos(self.beta)
               - yy * (cos(self.gamma) * sin(self.alpha) - cos(self.alpha) * sin(self.gamma) * sin(self.beta))
               + zz * (sin(self.gamma) * sin(self.alpha) + cos(self.gamma) * cos(self.alpha) * sin(self.beta))
               )
        yyy = (xx * cos(self.beta) * sin(self.alpha)
               + yy * (cos(self.gamma) * cos(self.alpha) + sin(self.gamma) * sin(self.alpha) * sin(self.beta))
               - zz * (cos(self.alpha) * sin(self.gamma) - cos(self.gamma) * sin(self.alpha) * sin(self.beta))
               )
        zzz = -xx * sin(self.beta) + yy * cos(self.beta) * sin(self.gamma) + zz * cos(self.gamma) * cos(self.beta)

        # Evaluate the ellipsoid function.
        return ((xxx ** 2 / ((self.a + boundary_on * self.ba) ** 2))
                + (yyy ** 2 / ((self.b + boundary_on * self.bb) ** 2))
                + (zzz ** 2 / ((self.c + boundary_on * self.bc) ** 2)) - 1)



import logging
from pathlib import Path
from sys import stdout

from vcams.helper import process_working_dir


def setup_main_logger(part_name: str, working_dir: str | Path,
                      display_log: bool = True, overwrite_logs: bool = True,
                      log_debug: bool = False) -> Path:
    """Set up the main logger for a *VoxelPart* object and return its file Path."""
    logger_name = 'vcams'  # This will become the root for all library-level logs.
    logger_obj = logging.getLogger(logger_name)
    filemode = 'w' if overwrite_logs else 'a'
    # Process and Validate working_dir.
    # This may be the second time that this is done, but it's OK.
    working_dir = process_working_dir(working_dir, part_name)
    log_file_path = Path(working_dir) / (part_name + '.log')  # Make sure it's necessary.

    log_level = logging.DEBUG if log_debug else logging.INFO

    # Create file handler and its format.
    file_handler_obj = logging.FileHandler(log_file_path, filemode)
    file_handler_formatter = logging.Formatter(fmt='%(asctime)s - %(levelname) 5s - %(message)s',
                                               datefmt='%Y-%m-%d %H:%M:%S')
    file_handler_obj.setFormatter(file_handler_formatter)
    logger_obj.addHandler(file_handler_obj)

    if display_log:
        stream_handler_obj = logging.StreamHandler(stream=stdout)
        stream_handler_formatter = logging.Formatter(fmt='%(asctime)s - %(levelname) 5s - %(message)s',
                                                     datefmt='%H:%M:%S')
        stream_handler_obj.setFormatter(stream_handler_formatter)
        logger_obj.addHandler(stream_handler_obj)
    logger_obj.setLevel(log_level)
    return log_file_path


def setup_dispersion_logger(part_name: str, log_file: Path, display_log: bool = False,
                            overwrite_logs: bool = True):
    """Set up the dispersion logger for a *ShapeDispersionArray* instance defined for a *VoxelPart* instance."""
    logger_name = part_name + '_dispersion_log'
    logger_obj = logging.getLogger(logger_name)
    filemode = 'w' if overwrite_logs else 'a'

    # Create file handler and its format.
    file_handler_obj = logging.FileHandler(log_file, filemode)
    file_handler_obj.terminator = ''
    file_handler_formatter = logging.Formatter(fmt='%(asctime)s - %(message)s', datefmt='%H:%M:%S')
    file_handler_obj.setFormatter(file_handler_formatter)
    logger_obj.addHandler(file_handler_obj)

    if display_log:
        stream_handler_obj = logging.StreamHandler(stream=stdout)
        stream_handler_obj.terminator = ''
        stream_handler_formatter = logging.Formatter(fmt='%(asctime)s - %(message)s', datefmt='%H:%M:%S')
        stream_handler_obj.setFormatter(stream_handler_formatter)
        logger_obj.addHandler(stream_handler_obj)
    logger_obj.setLevel(logging.DEBUG)

    # Log creation of the object.
    logger_obj.debug(f"Starting shape dispersion in part '{part_name}'.\n\n")
    return logger_obj


class LogWithoutFormatContext:
    """A context manager that allows logging to happen without any formatting.
    It saves each handler's formatter, sets it to an empty one and restores it when exiting.
    """
    # See https://docs.python.org/3/howto/logging-cookbook.html#using-a-context-manager-for-selective-logging
    bare_handler_formatter = logging.Formatter(fmt='')

    def __init__(self, logger_obj):  # , level=None, handler=None, close=True):
        self.logger_obj = logger_obj
        self.old_formatter_list = []

    def __enter__(self):
        for hndlr in self.logger_obj.handlers:
            self.old_formatter_list.append(hndlr.formatter)
            hndlr.setFormatter(self.bare_handler_formatter)

    def __exit__(self, et, ev, tb):
        for hndlr in self.logger_obj.handlers:
            hndlr.setFormatter(self.old_formatter_list.pop(0))



"""Various helper functions used throughout the library."""

import csv
from configparser import ConfigParser
from io import StringIO
from logging import getLogger
from pathlib import Path
from typing import Iterable

from numpy import unique
from numpy.typing import NDArray

from vcams.mask.image import image_thresh_mode_list
from vcams.mask.tpms import tpms_dict

logger = getLogger(__name__)


def is_name_valid(name: str) -> bool:
    # noinspection GrazieInspection
    """Check whether a string represents a valid name.
    Abaqus™ has many rules for names (labels) in the input files.
    The strictest combination is implemented here to ensure that
    a name is suitable for all purposes.

    This means that a name:

    + Must be 1-38 characters. This is because some object names
      in the abaqus scripting interface have a 38-character limit
      for their names.
    + May contain whitespace (if enclosed by double quotation marks).
    + Must start with a letter.
    + Must not begin or end with an underscore.
    + Must not include the following characters: ``$&*~!()[]{}|;'`",.?/\\``
    + Must not contain periods. This also means that any file names
      cannot contain any extensions. They will be added automatically.
    + Must be ASCII-compatible. This is checked by attempting :code:`str.decode('ascii')`
      and checking for :obj:`UnicodeDecodeError`.

    For more information, refer to:

    + *Labels* under the *Input Syntax Rules* section of the Abaqus Analysis User's Manual.
    + The documentation for the *InvalidNameError* object under
      the section *Standard Abaqus Scripting Interface Exceptions* of
      Abaqus Scripting User's Manual.

    Args:
        name: The string to be checked.

    Returns:
        Returns True for a valid name otherwise returns False.
    """
    # The regex ^(?=.*[ -~])(?=.*[^$&*~!()\[\]{}|;'`",.?/\\])(?=^[A-Za-z])^.{1,37}[^_]$
    # MAY be useful but the following method is easier to understand.
    forbidden_chars = "$&*~!()[]{}|;\'`\",.?/\\"
    if not isinstance(name, str):
        return False
    elif len(name) < 1:
        return False
    elif len(name) > 38:
        return False
    elif not name[0].isalpha():
        return False
    elif name.endswith('_'):  # The beginning is checked above.
        return False
    elif any((n in forbidden_chars) for n in name):
        return False
    elif not (name.isascii() and name.isprintable()):
        # Source: https://stackoverflow.com/a/51141941/7180705.
        return False
    else:
        return True


def return_default_working_dir(part_name: str = None) -> Path:
    """Return a suitable path in the user's Desktop for storing
    the intermediate and final results of the program.

    Args:
        part_name: Name of the part which is to be output
                   which must be valid according to :func:`is_name_valid`.
                   If set to *None*, the folder will be named :code:`results`.
                   Defaults to *None*.

    Returns:
        A path object containing the full path of a suitable folder in the users Desktop.
    """
    parts = ['Desktop', 'VCAMS Working Directory']
    # Validate part_name.
    if part_name is None:
        pass  # No subfolder.
    elif is_name_valid(part_name):
        parts.append(part_name)
    else:
        raise ValueError('part_name is not valid.')

    return Path.home().joinpath(*parts)


def process_working_dir(working_dir: Path | str | None, part_name: str) -> Path:
    """Process and validate the working_dir which represents
    the path where the intermediate and final results of the program are stored.
    This function validates the given *working_dir* parameter
    and if set to *None*, creates a default value using the :func:`return_default_working_dir` function.
    Then, the directory is created and finally a Path object referring to it is returned.

    Args:
        working_dir: Path to the folder where the final results, temporary file, and log files will be stored.
                     If set to *None* a suitable folder is automatically created in the user's home directory.
        part_name: Name of the part which is to be output
                   which must be valid according to :func:`is_name_valid`.
                   If set to *None*, the folder will be named :code:`results`.
                   Defaults to *None*.

    Returns:
        A path object containing the full path of a valid working directory that exists.
    """
    if working_dir is None:
        working_dir = return_default_working_dir(part_name)
    else:
        working_dir = Path(working_dir)
    working_dir.mkdir(parents=True, exist_ok=True)
    return working_dir


def write_to_logger_streams(msg: str):
    """Write a message directly to all handlers of the module's root logger object.

    Args:
        msg: The message that is written.
    """
    for handler in logger.root.handlers:
        handler.stream.writelines(msg)


def read_configuration(file_path: str) -> tuple[dict, dict, dict, dict]:
    """Read a configuration file containing all the information used for creating
    a :class:`~.voxelpart.VoxelPart` and return the information as a list of dictionaries.

    Args:
        file_path: Path to the configuration file.

    Returns:
        A tuple of the following dictionaries (part_creation_dict, part_manipulation_dict, bc_dict, output_dict).
    """
    # Read the config file.
    logger.debug('Reading configuration file at %s' % file_path)
    config = ConfigParser()
    config.read(file_path)

    # Check validity of the imported settings.
    section_list = ('Basic', 'Modeling', 'BC', 'Output')
    for name in section_list:
        if name not in config.sections():
            raise ValueError('Section "%s" was not present in the settings file.' % name)

    # Process various part of the config file.
    # Basic: Information used for creating the part.
    part_creation_dict = dict()
    basic_section = config['Basic']
    dim = basic_section['dim'].upper()
    if dim == '2D':
        part_creation_dict['size'] = (int(basic_section['num_voxels_x']),
                                      int(basic_section['num_voxels_y']))
    elif dim == '3D':
        part_creation_dict['size'] = (int(basic_section['num_voxels_x']),
                                      int(basic_section['num_voxels_y']),
                                      int(basic_section['num_voxels_z']))
    else:
        raise ValueError('Field "dim" is set to %s, which is invalid.' % basic_section['dim'])
    if 'base_material' in basic_section:
        part_creation_dict['base_material'] = int(basic_section['base_material'])
    else:
        part_creation_dict['base_material'] = 0
    if dim == '2D':
        part_creation_dict['voxel_size'] = (float(basic_section['voxel_size_x']),
                                            float(basic_section['voxel_size_y']))
    elif dim == '3D':
        part_creation_dict['voxel_size'] = (float(basic_section['voxel_size_x']),
                                            float(basic_section['voxel_size_y']),
                                            float(basic_section['voxel_size_z']))
    else:
        raise ValueError('Field "dim" is set to %s, which is invalid.' % basic_section['dim'])
    num_mats = basic_section['num_mats']
    if num_mats == '0':
        part_creation_dict['dtype'] = 'uint8'
    elif num_mats == '1':
        part_creation_dict['dtype'] = 'uint16'
    elif num_mats == '2':
        part_creation_dict['dtype'] = 'uint32'
    elif num_mats == '3':
        part_creation_dict['dtype'] = 'uint54'
    else:
        raise ValueError("Invalid value for field 'num_mats'.")
    part_creation_dict['name'] = basic_section['part_name']
    part_creation_dict['description'] = basic_section['part_description']
    part_creation_dict['working_dir'] = basic_section['working_dir']
    part_creation_dict['overwrite_logs'] = True
    part_creation_dict['log_debug'] = config.getboolean('Basic', 'log_debug')

    # Modeling: Manipulating the part.
    part_manipulation_dict = dict()
    modeling_section = config['Modeling']
    modeling_mode = modeling_section['modeling_mode']
    part_manipulation_dict['modeling_mode'] = modeling_mode
    part_manipulation_dict['dim'] = dim
    if modeling_mode == '0':  # No action selected. This is technically invalid, but is considered to be '1'.
        pass
    elif modeling_mode == '1':  # No Further Manipulation.
        pass
    elif modeling_mode == '2':  # Random Element Dispersion
        part_manipulation_dict['random_phase_fraction'] = float(modeling_section['random_phase_fraction'])
        part_manipulation_dict['random_phase_matcode'] = int(modeling_section['random_phase_matcode'])
    elif modeling_mode == '3':  # TPMS
        tpms_type = modeling_section['tpms_type']
        if int(tpms_type) in tpms_dict.keys():
            part_manipulation_dict['tpms_type'] = int(tpms_type)
        else:
            raise ValueError('Field "tpms_type" is set to %s, which is invalid.' % tpms_type)
        part_manipulation_dict['tpms_length'] = float(modeling_section['tpms_length'])
        part_manipulation_dict['tpms_constant'] = float(modeling_section['tpms_constant'])
        part_manipulation_dict['tpms_fill_value'] = int(modeling_section['tpms_fill_value'])
    elif modeling_mode == '4':  # Image Processing (Single 2D Image)
        part_manipulation_dict['single_image_path'] = modeling_section['single_image_path']
        part_manipulation_dict['single_image_scale'] = float(modeling_section['single_image_scale'])
        # Make sure the following is identical for single and multiple image modeling modes.
        image_thresh_key = 'single_image_thresh_mode'
        image_thresh_mode = modeling_section[image_thresh_key].lower()
        if image_thresh_mode in image_thresh_mode_list:
            part_manipulation_dict[image_thresh_key] = image_thresh_mode
        else:
            raise ValueError("Invalid value for field '%s'." % image_thresh_key)
        part_manipulation_dict['single_image_thresh_value'] = float(modeling_section['single_image_thresh_value'])
        part_manipulation_dict['single_image_denoise'] = config.getboolean('Modeling', 'single_image_denoise')
    elif modeling_mode == '5':  # Stack of 2D images for a 3D part.
        part_manipulation_dict['multi_image_path'] = modeling_section['multi_image_path']
        part_manipulation_dict['multi_image_scale'] = float(modeling_section['multi_image_scale'])
        # Make sure the following is identical for single and multiple image modeling modes.
        image_thresh_key = 'multi_image_thresh_mode'
        image_thresh_mode = modeling_section[image_thresh_key].lower()
        if image_thresh_mode in image_thresh_mode_list:
            part_manipulation_dict[image_thresh_key] = image_thresh_mode
        else:
            raise ValueError("Invalid value for field '%s'." % image_thresh_key)
        part_manipulation_dict['multi_image_thresh_value'] = float(modeling_section['multi_image_thresh_value'])
        part_manipulation_dict['multi_image_denoise'] = config.getboolean('Modeling', 'multi_image_denoise')
    elif modeling_mode == '6':  # Planar Composite (Circular Inclusions)
        part_manipulation_dict['circle_list'] = csv_string_to_list(modeling_section['modeling_circle_table'])
    elif modeling_mode == '7':  # Spatial Composite (Spherical Inclusions)
        part_manipulation_dict['sphere_list'] = csv_string_to_list(modeling_section['modeling_sphere_table'])
    else:
        raise ValueError('Field "modeling_mode" is set to %s, which is invalid.' % modeling_mode)

    # BC: Boundary Conditions.
    bc_dict = dict()
    bc_section = config['BC']
    bc_dict['dim'] = dim
    bc_type = bc_section['bc_type']
    bc_dict['bc_type'] = bc_type  # There are no parameters associated with BCs.

    # Output.
    output_dict = dict()
    output_section = config['Output']
    if is_name_valid(name=output_section['file_name']):
        output_dict['file_name'] = output_section['file_name']
    else:
        raise ValueError('Field "file_name" contains an invalid name.')
    output_dict['elem_code'] = output_section['elem_code']
    output_dict['dim'] = dim
    output_mats_type = output_section['output_mats_type']
    if output_mats_type == '0':  # All Materials.
        output_dict['material_elem_sets'] = 'All'
    elif output_mats_type == '1':  # Non-Empty Materials.
        output_dict['material_elem_sets'] = 'Non-Empty'
    elif output_mats_type == '2':  # Output Selected Materials.
        output_dict['material_elem_sets'] = [int(i) for i in output_section['output_mats_select'].split(',')]
    else:
        raise ValueError('Field "output_mats_type" is set to %s, which is invalid.' % output_mats_type)

    logger.debug('Configuration file read successfully.')
    return part_creation_dict, part_manipulation_dict, bc_dict, output_dict


def csv_string_to_list(csv_string: str) -> list:
    """Convert a csv string to a list. The csv dialect is sniffed using Python's *csv* module.

    Args:
        csv_string: A string of comma separated values created using Python's *csv* module.

    Returns:
        List of rows detected in the csv string.
    """
    buffer_io = StringIO(csv_string)
    dialect = csv.Sniffer().sniff(buffer_io.readline())
    buffer_io.seek(0)
    csv_reader = csv.reader(buffer_io, dialect)
    return [row for row in csv_reader]


def validate_dim(dim: str):
    """Validate the *dim* parameter used throughout the library.
    Returns:
        ValueError: If dim is not '2D' or '3D'.
    """
    if dim.upper() not in ['2D', '3D']:
        raise ValueError("dim can only be one of '2D' or '3D'.")


def validate_materials_to_be_output(part, material_list: str | int | Iterable | NDArray):
    """Validate the list of materials that are to be output and return a list of material codes.

    Args:
        part (VoxelPart): The *VoxelPart* instance for which the operation is performed.
        material_list: One of the following:

          + *'All'* which outputs all materials in *part*.
          + *'Non-Empty'* which outputs all non-zero (=non-empty) materials in *part*.
          + An integer or a tuple of integer material codes corresponding
            to the materials that should be written to the output.
            All material codes should exist in *part*.

    Returns:
        A list of valid integer material codes that have been selected.
    """
    # Validate and process material_elem_sets.
    valid_materials = unique(part.data)
    if isinstance(material_list, str):
        if material_list.upper() in ['ALL', 'NON-EMPTY']:
            selected_mats = list(valid_materials)
            if (0 in selected_mats) and (material_list.upper() == 'NON-EMPTY'):
                selected_mats.remove(0)
            material_list = selected_mats
        else:
            raise ValueError(f"Invalid string '{material_list}' for material_list."
                             f"Valid values are 'All' and 'Non-Empty'.")
    else:  # Will raise most errors.
        for mat in material_list:
            if mat not in valid_materials:
                raise ValueError(f'Material {mat} specified in material_list is not present in the model.')
    return material_list
